from typing import List
from typing import Optional

from tecton import conf
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import sdk_public_method
from tecton.fco import Fco
from tecton_proto.data import entity_pb2
from tecton_proto.metadataservice.metadata_service_pb2 import GetEntityRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetEntitySummaryRequest
from tecton_spark.fco_container import FcoContainer
from tecton_spark.id_helper import IdHelper
from tecton_spark.logger import get_logger

logger = get_logger("Entity")


class Entity(Fco):
    """
    Entity class.

    An Entity is a class that represents an Entity that is being modelled in Tecton.
    Entities are used to index and organize features - a :class:`FeatureView`
    contains at least one Entity.

    Entities contain metadata about *join keys*, which represent the columns
    that are used to join features together.
    """

    _entity_proto: entity_pb2.Entity = None
    _fco_container: FcoContainer

    def __init__(self):
        """Do not call this directly. Use :py:func:`tecton.get_entity`"""

    @classmethod
    def _from_proto(cls, entity_proto: entity_pb2.Entity, fco_container: FcoContainer) -> "Entity":
        """
        Instantiates a new Entity object.

        `Entity` instance always corresponds to the existing registered Entity in the Database.
        """
        obj = cls.__new__(cls)
        obj._entity_proto = entity_proto
        obj._fco_container = fco_container
        return obj

    @classmethod
    def _fco_type_name_singular_snake_case(cls) -> str:
        return "entity"

    @classmethod
    def _fco_type_name_plural_snake_case(cls) -> str:
        return "entities"

    @property
    def _fco_metadata(self):
        return self._entity_proto.fco_metadata

    def __repr__(self):
        return f"Entity(name='{self.name}', join_keys={self.join_keys}, description='{self.description}')"

    @property  # type: ignore
    @sdk_public_method
    def join_keys(self) -> List[str]:
        """
        Returns the join keys of this entity.
        """
        return list(self._entity_proto.join_keys)

    @sdk_public_method
    def summary(self) -> Displayable:
        """
        Displays a human readable summary of this Entity.
        """
        request = GetEntitySummaryRequest()
        request.fco_locator.id.CopyFrom(self._entity_proto.entity_id)
        request.fco_locator.workspace = self.workspace

        response = metadata_service.instance().GetEntitySummary(request)
        return Displayable.from_fco_summary(response.fco_summary)

    @property
    def id(self) -> str:
        """
        Returns ID of the database entity entry.
        """
        return IdHelper.to_string(self._entity_proto.entity_id)


@sdk_public_method
def get_entity(name: str, workspace_name: Optional[str] = None) -> Entity:
    """
    Fetch an existing :class:`Entity` by name.

    :param name: Unique name for an existing entity.

    :raises TectonValidationError: if an entity with given name does not exist.
    """

    if workspace_name == None:
        logger.warning(
            "`tecton.get_entity('<name>')` is deprecated. Please use `tecton.get_workspace('<workspace_name>').get_entity('<name>')` instead."
        )
    request = GetEntityRequest()
    request.name = name
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    request.disable_legacy_response = True
    response = metadata_service.instance().GetEntity(request)

    fco_container = FcoContainer(response.fco_container)
    entity_proto = fco_container.get_single_root()

    if not entity_proto:
        raise errors.FCO_NOT_FOUND(Entity, name)

    return Entity._from_proto(entity_proto, fco_container)



from typing import Optional
from typing import Union

import pandas as pd
from pyspark.sql import DataFrame

import tecton.interactive.data_frame
from tecton import conf
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import sdk_public_method
from tecton.fco import Fco
from tecton.interactive import data_frame
from tecton.tecton_context import TectonContext
from tecton_proto.args.new_transformation_pb2 import TransformationMode
from tecton_proto.data.new_transformation_pb2 import NewTransformation as TransformationProto
from tecton_proto.metadataservice.metadata_service_pb2 import GetTransformationRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetTransformationSummaryRequest
from tecton_spark import function_serialization as func_ser
from tecton_spark.fco_container import FcoContainer
from tecton_spark.logger import get_logger
from tecton_spark.materialization_context import BaseMaterializationContext

logger = get_logger("Transformation")
TRANSFORMATION_TEMP_VIEW_PREFIX = "_tecton_transformation_run_"
CONST = Union[str, int, float, bool]


class Transformation(Fco):
    """
    Transformation Class.

    A Transformation is a Tecton Object that contains logic for creating a feature.

    To get a Transformation instance, call :py:func:`tecton.get_transformation`.
    """

    _transformation_proto: TransformationProto
    _fco_container: FcoContainer

    def __init__(self):
        """Do not call this directly. Use :py:func:`tecton.get_transformation`"""

    @classmethod
    def _fco_type_name_singular_snake_case(cls) -> str:
        return "transformation"

    @classmethod
    def _fco_type_name_plural_snake_case(cls) -> str:
        return "transformations"

    @property
    def _fco_metadata(self):
        return self._transformation_proto.fco_metadata

    @property
    def transformer(self):
        """
        Returns the raw transformation function encapsulated by this Transformation.
        """
        return func_ser.from_proto(self._transformation_proto.user_function, scope={})

    @classmethod
    def _from_proto(cls, transformation: TransformationProto, fco_container: FcoContainer):
        """
        Returns a Transformation instance.

        :param transformations: Transformation proto object.
        """
        obj = Transformation.__new__(cls)
        obj._transformation_proto = transformation
        obj._fco_container = fco_container
        return obj

    @sdk_public_method
    def run(
        self,
        *inputs: Union["pd.DataFrame", "pd.Series", "DataFrame", CONST],
        context: BaseMaterializationContext = None,
    ) -> data_frame.DataFrame:
        """Run the transformation against inputs.

        :param inputs: positional arguments to the transformation function. For PySpark and SQL transformations,
                       these are either ``pandas.DataFrame`` or ``pyspark.sql.DataFrame`` objects.
                       For on-demand transformations, these are ``pandas.Dataframe`` objects.
        :param context: (Optional) An optional materialization context object.
        """

        if self._transformation_proto.transformation_mode == TransformationMode.TRANSFORMATION_MODE_SPARK_SQL:
            return self._sql_run(*inputs, context=context)
        elif self._transformation_proto.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PYSPARK:
            return self._pyspark_run(*inputs, context=context)
        elif self._transformation_proto.transformation_mode == TransformationMode.TRANSFORMATION_MODE_PANDAS:
            return self._on_demand_run(*inputs)
        raise RuntimeError(f"{self._transformation_proto.transformation_mode} does not support `run(...)`")

    def _sql_run(self, *inputs, context) -> data_frame.DataFrame:
        def create_temp_view(df, dataframe_index):
            df = tecton.interactive.data_frame.DataFrame._create(df).to_spark()
            temp_view = f"{TRANSFORMATION_TEMP_VIEW_PREFIX}{self._fco_metadata.name}_input_{dataframe_index}"
            df.createOrReplaceTempView(temp_view)
            return temp_view

        args = [create_temp_view(v, i) if not isinstance(v, CONST.__args__) else v for i, v in enumerate(inputs)]
        if context is not None:
            args.append(context)

        spark = TectonContext.get_instance()._get_spark()
        return tecton.interactive.data_frame.DataFrame._create(spark.sql(self.transformer(*args)))

    def _pyspark_run(self, *inputs, context) -> data_frame.DataFrame:
        args = [
            tecton.interactive.data_frame.DataFrame._create(v).to_spark() if not isinstance(v, CONST.__args__) else v
            for v in inputs
        ]
        if context is not None:
            args.append(context)

        return tecton.interactive.data_frame.DataFrame._create(self.transformer(*args))

    def _on_demand_run(self, *inputs) -> data_frame.DataFrame:
        for df in inputs:
            if not isinstance(df, pd.DataFrame):
                raise TypeError(f"Input must be of type pandas.DataFrame, but was {type(df)}.")

        return tecton.interactive.data_frame.DataFrame._create(self.transformer(*inputs))

    def summary(self):
        """
        Displays a human readable summary of this Transformation.
        """
        request = GetTransformationSummaryRequest()
        request.fco_locator.id.CopyFrom(self._transformation_proto.transformation_id)
        request.fco_locator.workspace = self.workspace

        response = metadata_service.instance().GetTransformationSummary(request)
        return Displayable.from_fco_summary(response.fco_summary)


@sdk_public_method
def get_transformation(name, workspace_name: Optional[str] = None) -> Transformation:
    """
    Fetch an existing :class:`Transformation` by name.

    :param name: An unique name of the registered Transformation.

    :return: A :class:`Transformation` class instance.

    :raises TectonValidationError: if a Transformation with the passed name is not found.
    """
    if workspace_name == None:
        logger.warning(
            "`tecton.get_transformation('<name>')` is deprecated. Please use `tecton.get_workspace('<workspace_name>').get_transformation('<name>')` instead."
        )

    request = GetTransformationRequest()
    request.name = name
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    request.disable_legacy_response = True

    response = metadata_service.instance().GetTransformation(request)
    fco_container = FcoContainer(response.fco_container)
    transformation_proto = fco_container.get_single_root()

    if not transformation_proto:
        raise errors.FCO_NOT_FOUND(Transformation, name)

    return Transformation._from_proto(transformation_proto, fco_container)



from typing import List
from typing import Optional

from tecton import conf
from tecton._internals import metadata_service
from tecton._internals.sdk_decorators import sdk_public_method
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllEntitiesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllFeatureServicesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllSavedFeatureDataFramesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllTransformationsRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetAllVirtualDataSourcesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import ListWorkspacesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import QueryFeatureViewsRequest
from tecton_spark.fco_container import FcoContainer


@sdk_public_method
def list_feature_views(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of all registered FeatureViews.

    :return: A list of strings.
    """
    request = QueryFeatureViewsRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    request.disable_legacy_response = True
    response = metadata_service.instance().QueryFeatureViews(request)
    fco_container = FcoContainer(response.fco_container)
    return sorted(
        [proto.fco_metadata.name for proto in fco_container.get_root_fcos() if not proto.HasField("feature_table")]
    )


@sdk_public_method
def list_feature_tables(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of all registered FeatureTables.

    :return: A list of strings.
    """
    request = QueryFeatureViewsRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    request.disable_legacy_response = True
    response = metadata_service.instance().QueryFeatureViews(request)
    fco_container = FcoContainer(response.fco_container)
    return sorted(
        [proto.fco_metadata.name for proto in fco_container.get_root_fcos() if proto.HasField("feature_table")]
    )


@sdk_public_method
def list_feature_services(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of all registered FeatureServices.

    :return: A list of strings.
    """
    request = GetAllFeatureServicesRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetAllFeatureServices(request)
    return sorted([proto.fco_metadata.name for proto in response.feature_services])


@sdk_public_method
def list_transformations(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of all registered Transformations.

    :return: A list of strings.
    """
    request = GetAllTransformationsRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetAllTransformations(request)
    return sorted([proto.fco_metadata.name for proto in response.new_transformations])


@sdk_public_method
def list_entities(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of all registered Entities.

    :returns: A list of strings.
    """
    request = GetAllEntitiesRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetAllEntities(request)
    return sorted([proto.fco_metadata.name for proto in response.entities])


@sdk_public_method
def list_data_sources(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of the names of all registered Data Sources.

    :return: A list of strings.
    """
    request = GetAllVirtualDataSourcesRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetAllVirtualDataSources(request)
    return sorted([proto.fco_metadata.name for proto in response.virtual_data_sources])


@sdk_public_method
def list_workspaces() -> List[str]:
    """
    Returns a list of the names of all registered Workspaces.

    :return: A list of strings.
    """
    request = ListWorkspacesRequest()
    response = metadata_service.instance().ListWorkspaces(request)
    return sorted([workspace.name for workspace in response.workspaces])


@sdk_public_method
def list_datasets(workspace_name: Optional[str] = None) -> List[str]:
    """
    Returns a list of all registered Datasets.

    :return: A list of strings.
    """
    request = GetAllSavedFeatureDataFramesRequest()
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetAllSavedFeatureDataFrames(request)
    return sorted([sfdf.info.name for sfdf in response.saved_feature_dataframes])



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protoc-gen-swagger/options/annotations.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
from protoc_gen_swagger.options import openapiv2_pb2 as protoc__gen__swagger_dot_options_dot_openapiv2__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protoc-gen-swagger/options/annotations.proto',
  package='grpc.gateway.protoc_gen_swagger.options',
  syntax='proto3',
  serialized_options=b'ZAgithub.com/grpc-ecosystem/grpc-gateway/protoc-gen-swagger/options',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,protoc-gen-swagger/options/annotations.proto\x12\'grpc.gateway.protoc_gen_swagger.options\x1a google/protobuf/descriptor.proto\x1a*protoc-gen-swagger/options/openapiv2.proto:|\n\x11openapiv2_swagger\x12\x1c.google.protobuf.FileOptions\x18\x92\x08 \x01(\x0b\x32\x30.grpc.gateway.protoc_gen_swagger.options.SwaggerR\x10openapiv2Swagger:\x84\x01\n\x13openapiv2_operation\x12\x1e.google.protobuf.MethodOptions\x18\x92\x08 \x01(\x0b\x32\x32.grpc.gateway.protoc_gen_swagger.options.OperationR\x12openapiv2Operation:|\n\x10openapiv2_schema\x12\x1f.google.protobuf.MessageOptions\x18\x92\x08 \x01(\x0b\x32/.grpc.gateway.protoc_gen_swagger.options.SchemaR\x0fopenapiv2Schema:s\n\ropenapiv2_tag\x12\x1f.google.protobuf.ServiceOptions\x18\x92\x08 \x01(\x0b\x32,.grpc.gateway.protoc_gen_swagger.options.TagR\x0copenapiv2Tag:|\n\x0fopenapiv2_field\x12\x1d.google.protobuf.FieldOptions\x18\x92\x08 \x01(\x0b\x32\x33.grpc.gateway.protoc_gen_swagger.options.JSONSchemaR\x0eopenapiv2FieldBCZAgithub.com/grpc-ecosystem/grpc-gateway/protoc-gen-swagger/optionsb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,protoc__gen__swagger_dot_options_dot_openapiv2__pb2.DESCRIPTOR,])


OPENAPIV2_SWAGGER_FIELD_NUMBER = 1042
openapiv2_swagger = _descriptor.FieldDescriptor(
  name='openapiv2_swagger', full_name='grpc.gateway.protoc_gen_swagger.options.openapiv2_swagger', index=0,
  number=1042, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='openapiv2Swagger', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
OPENAPIV2_OPERATION_FIELD_NUMBER = 1042
openapiv2_operation = _descriptor.FieldDescriptor(
  name='openapiv2_operation', full_name='grpc.gateway.protoc_gen_swagger.options.openapiv2_operation', index=1,
  number=1042, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='openapiv2Operation', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
OPENAPIV2_SCHEMA_FIELD_NUMBER = 1042
openapiv2_schema = _descriptor.FieldDescriptor(
  name='openapiv2_schema', full_name='grpc.gateway.protoc_gen_swagger.options.openapiv2_schema', index=2,
  number=1042, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='openapiv2Schema', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
OPENAPIV2_TAG_FIELD_NUMBER = 1042
openapiv2_tag = _descriptor.FieldDescriptor(
  name='openapiv2_tag', full_name='grpc.gateway.protoc_gen_swagger.options.openapiv2_tag', index=3,
  number=1042, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='openapiv2Tag', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
OPENAPIV2_FIELD_FIELD_NUMBER = 1042
openapiv2_field = _descriptor.FieldDescriptor(
  name='openapiv2_field', full_name='grpc.gateway.protoc_gen_swagger.options.openapiv2_field', index=4,
  number=1042, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, json_name='openapiv2Field', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)

DESCRIPTOR.extensions_by_name['openapiv2_swagger'] = openapiv2_swagger
DESCRIPTOR.extensions_by_name['openapiv2_operation'] = openapiv2_operation
DESCRIPTOR.extensions_by_name['openapiv2_schema'] = openapiv2_schema
DESCRIPTOR.extensions_by_name['openapiv2_tag'] = openapiv2_tag
DESCRIPTOR.extensions_by_name['openapiv2_field'] = openapiv2_field
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

openapiv2_swagger.message_type = protoc__gen__swagger_dot_options_dot_openapiv2__pb2._SWAGGER
google_dot_protobuf_dot_descriptor__pb2.FileOptions.RegisterExtension(openapiv2_swagger)
openapiv2_operation.message_type = protoc__gen__swagger_dot_options_dot_openapiv2__pb2._OPERATION
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(openapiv2_operation)
openapiv2_schema.message_type = protoc__gen__swagger_dot_options_dot_openapiv2__pb2._SCHEMA
google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(openapiv2_schema)
openapiv2_tag.message_type = protoc__gen__swagger_dot_options_dot_openapiv2__pb2._TAG
google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(openapiv2_tag)
openapiv2_field.message_type = protoc__gen__swagger_dot_options_dot_openapiv2__pb2._JSONSCHEMA
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(openapiv2_field)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



"""
Functions for serializing Python functions.

The only known areas where we will fail to serialize something usable are:

user imports like import foo.bar.baz instead of from foo.bar import baz
are not detected properly by the underlying library we use for looking up
dependencies. import json or other top-level modeuls still works just fine,
its just when you include multiple path components that we don't detect the
dependency properly

references to instantiated non-builtin types with a repr that can't be exec'd

users referring to a renamed version of an imported module has some wonky
behavior, for example `import json` then `foo=json` then referring to
`foo` in the func to be serialized.
"""
import inspect
from collections import defaultdict
from types import FunctionType
from types import ModuleType

from dill.detect import freevars
from dill.detect import globalvars

from tecton_proto.args.user_defined_function_pb2 import UserDefinedFunction
from tecton_spark.errors import TectonValidationError
from tecton_spark.logger import get_logger
from tecton_spark.repo_file_handler import ensure_prepare_repo
from tecton_spark.repo_file_handler import repo_files_set
from tecton_spark.time_utils import WINDOW_UNBOUNDED_PRECEDING

logger = get_logger("function_serialization")


def to_proto(transform):
    code = transform._code if hasattr(transform, "_code") else _getsource(transform)
    _validate_code(code)

    proto = UserDefinedFunction()
    proto.name = transform.__name__
    proto.body = code

    return proto


BANNED_INDIRECT_IMPORTS = ["materialization_context", "tecton_sliding_window_udf", "WINDOW_UNBOUNDED_PRECEDING"]


def _validate_code(code):
    for banned_import in BANNED_INDIRECT_IMPORTS:
        if f"tecton.{banned_import}" in code:
            raise Exception(
                f"Cannot serialize `tecton.{banned_import}`. Please use an import like `from tecton import {banned_import}`. Note that even comments may trigger this if your code contains `tecton.{banned_import}`"
            )


# If these are in the path to a module, we don't expect the function definitions to be inside a feature repo, but we still want to inline them
# integration_tests/ is added as a hack because our integration tests aren't run in a feature repo. But unless a customer is importing from a python package that has a directory named integration_tests, it
# shouldn't cause any wrong behavior.
TECTON_TRANSFORMATION_WHITELIST = ["builtin_transformations.py", "integration_tests/"]

# returns true if the module is defined by tecton, rather than by the user
def is_tecton_defined(module):
    return sum([whitelist_str in module for whitelist_str in TECTON_TRANSFORMATION_WHITELIST]) > 0


# UNDONE:
# this breaks for the following case:
# def f():
#    a = module.CONSTANT
#
# for some reason, the module/constant don't show up in globalvars.
# this case has not occurred in practice
def _getsource(func):
    imports = defaultdict(set)
    modules = set()
    code_lines = []
    seen_args = {}

    # file the func was defined in
    func_file = inspect.getfile(inspect.getmodule(func))

    # Check if the file is part of the tecton codebase, not the feature repo.
    # Only perform any feature repo initialization for functions in the feature repo.
    # We short-circuit any repo_file_handler functions that depend on initialization if we skip initialization here.
    if not is_tecton_defined(func_file):
        ensure_prepare_repo(func_file)

    def process_functiontype(name, obj, imports, modules, code_lines, seen_args):
        # if this is user-defined or otherwise unavailable to import at deserialization time
        # including anything without a module, with module of __main__, defined inside the feature repo
        # or anything with a qualified name containing characters invalid for an import path, like foo.<locals>.bar
        module = inspect.getfile(inspect.getmodule(obj))
        if (
            obj.__module__ in ("__main__", None)
            or "<" in obj.__qualname__
            or is_tecton_defined(module)
            or module in repo_files_set()
        ):
            objs = globalvars(obj, recurse=False)
            objs.update(freevars(obj))
            default_objs = {}
            for param in inspect.signature(obj).parameters.values():
                if param.default != inspect.Parameter.empty:
                    default_objs[param.name] = param.default
                    # these defaults shadow over obj
                    objs.pop(param.name, None)
            # need to sort the keys since globalvars ordering is non-deterministic
            for dependency, dep_obj in sorted(objs.items()):
                recurse(dependency, dep_obj, imports, modules, code_lines, seen_args, write_codelines=True)
            for dependency, dep_obj in sorted(default_objs.items()):
                # we dont re-write defaults at top-level since the `def` lines should have the declarations
                recurse(dependency, dep_obj, imports, modules, code_lines, seen_args, write_codelines=False)
            fdef = inspect.getsource(obj)
            fdef = fdef[fdef.find("def ") :]
            code_lines.append(fdef)
        else:
            imports[obj.__module__].add(obj.__name__)

    def recurse(name, obj, imports, modules, code_lines, seen_args, write_codelines):
        def _add_codeline(line):
            if write_codelines:
                code_lines.append(line)

        # prevent processing same dependency object multiple times, even if
        # multiple dependent objects exist in the tree from the original
        # func
        seen_key = str(name) + str(obj)
        if seen_args.get(seen_key) is True:
            return
        seen_args[seen_key] = True

        has_spark = True
        try:
            import pyspark
        except ImportError:
            has_spark = False

        is_spark_struct = False
        is_unbound_materialization_context = False
        is_tecton_sliding_window_udf = False
        if has_spark:
            from pyspark.sql import types as spark_types

            is_spark_struct = isinstance(obj, spark_types.StructType)
            from tecton_spark.materialization_context import UnboundMaterializationContext

            is_unbound_materialization_context = isinstance(obj, UnboundMaterializationContext)

            from tecton_spark.udfs import tecton_sliding_window_udf

            is_tecton_sliding_window_udf = obj is tecton_sliding_window_udf

        # Confusingly classes are subtypes of 'type'; non-classes are not
        if isinstance(obj, type):
            if obj.__module__ == "__main__":
                raise Exception(f"Cannot serialize class {obj.__name__} from module __main__")
            imports[obj.__module__].add(obj.__name__)

        elif isinstance(obj, FunctionType):
            process_functiontype(name, obj, imports, modules, code_lines, seen_args)
        elif isinstance(obj, ModuleType):
            module_file = inspect.getfile(obj)
            # skip this check for functions written by Tecton
            if not is_tecton_defined(func_file) and module_file in repo_files_set():
                raise Exception(
                    f"Cannot serialize usage of {obj.__name__}. Instead, directly import objects from the module, e.g. `from {obj.__name__} import obj`"
                )
            if f"{obj.__package__}.{name}" == obj.__name__:
                imports[obj.__package__].add(name)
            else:
                modules.add(obj.__name__)
        elif is_spark_struct:
            _add_codeline(f"{name} = StructType.fromJson(json.loads('{obj.json()}'))")
            modules.add("json")
            imports["pyspark.sql.types"].add("StructType")
        elif is_unbound_materialization_context:
            imports["tecton_spark.materialization_context"].add("materialization_context")
        elif is_tecton_sliding_window_udf:
            imports["tecton_spark.udfs"].add("tecton_sliding_window_udf")
        elif obj == WINDOW_UNBOUNDED_PRECEDING:
            imports["tecton_spark.time_utils"].add("WINDOW_UNBOUNDED_PRECEDING")
        else:
            try:
                repr_str = f"{name}={repr(obj)}"
                exec(repr_str)
                _add_codeline(repr_str)
            except Exception:
                raise Exception(f"Cannot evaluate object {obj} of type '{type(obj)}' for serialization")

    recurse(func.__name__, func, imports, modules, code_lines, seen_args, write_codelines=True)

    for module in sorted(imports):
        import_line = f"from {module} import "
        import_line += ", ".join(sorted(imports[module]))
        code_lines.insert(0, import_line)

    for module in sorted(modules):
        code_lines.insert(0, f"import {module}")

    return "\n".join(code_lines)


def inlined(func):
    # this a no-op that can be deleted later
    return func


def from_proto(serialized_transform: UserDefinedFunction, scope=None):
    """
    deserialize into global scope by default. if a scope if provided, deserialize into provided scope
    """

    if scope is None:
        # PySpark has issues if the UDFs are not in global scope
        scope = __import__("__main__").__dict__

    assert serialized_transform.HasField("body") and serialized_transform.HasField(
        "name"
    ), "Invalid UserDefinedFunction."

    try:
        exec(serialized_transform.body, scope)
    except NameError as e:
        raise TectonValidationError(
            "Failed to serialize function. Please note that all imports must be in the body of the function (not top-level) and type annotations cannot require imports. Additionally, be cautious of variables that shadow other variables. See https://docs.tecton.ai/v2/overviews/framework/transformations.html for more details.",
            e,
        )

    # Return function pointer
    try:
        fn = eval(serialized_transform.name, scope)
        fn._code = serialized_transform.body
        return fn
    except Exception as e:
        raise ValueError("Invalid transform") from e



from base64 import b64encode
from datetime import datetime
from datetime import timezone
from typing import List
from typing import Optional
from typing import Union

import pandas as pd
import pendulum
from pyspark.sql import DataFrame as pysparkDF
from pyspark.sql.functions import array
from pyspark.sql.functions import udf
from pyspark.sql.types import ArrayType
from pyspark.sql.types import DoubleType
from pyspark.sql.types import FloatType
from pyspark.sql.types import IntegerType
from pyspark.sql.types import LongType
from pyspark.sql.types import StringType
from pyspark.sql.types import StructType

from tecton._internals import data_frame_helper
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals import utils
from tecton._internals.display import Displayable
from tecton._internals.feature_retrieval_internal import find_dependent_feature_set_items
from tecton._internals.feature_retrieval_internal import get_features
from tecton._internals.sdk_decorators import sdk_public_method
from tecton.fco import Fco
from tecton.interactive.data_frame import DataFrame
from tecton.interactive.dataset import Dataset
from tecton.tecton_context import TectonContext
from tecton_proto.data.materialization_status_pb2 import DataSourceType
from tecton_proto.data.materialization_status_pb2 import MaterializationStatus
from tecton_proto.metadataservice.metadata_service_pb2 import DeleteEntitiesRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetDeleteEntitiesInfoRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetFeatureViewSummaryRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetMaterializationStatusRequest
from tecton_proto.online_store.feature_value_pb2 import FeatureValueList
from tecton_proto.online_store.feature_value_pb2 import NullValue
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_spark.feature_set_config import FeatureSetConfig
from tecton_spark.id_helper import IdHelper
from tecton_spark.logger import get_logger
from tecton_spark.online_serving_index import OnlineServingIndex
from tecton_spark.schema import Schema
from tecton_spark.schema_spark_utils import column_name_spark_data_types
from tecton_spark.schema_spark_utils import schema_from_spark

logger = get_logger("FeatureDefinition")


class FeatureDefinition(Fco):
    @property
    def _fco_metadata(self):
        return self._proto.fco_metadata

    @property
    def _view_schema(self):
        return Schema(self._proto.schemas.view_schema)

    @property
    def _materialization_schema(self):
        return Schema(self._proto.schemas.materialization_schema)

    @property
    def _id_proto(self):
        return self._proto.feature_view_id

    @property  # type: ignore
    @sdk_public_method
    def id(self) -> str:
        """
        Returns the id of this object
        """
        return IdHelper.to_string(self._id_proto)

    @property
    def join_keys(self) -> List[str]:
        """
        Returns the join key column names
        """
        return list(self._proto.join_keys)

    @property  # type: ignore
    @sdk_public_method
    def online_serving_index(self) -> OnlineServingIndex:
        """
        Returns Defines the set of join keys that will be indexed and queryable during online serving.
        Defaults to the complete join key.
        """
        return OnlineServingIndex.from_proto(self._proto.online_serving_index)

    @property
    def wildcard_join_key(self) -> Optional[str]:
        """
        Returns a wildcard join key column name if it exists;
        Otherwise returns None.
        """
        online_serving_index = self.online_serving_index
        wildcard_keys = [join_key for join_key in self.join_keys if join_key not in online_serving_index.join_keys]
        return wildcard_keys[0] if wildcard_keys else None

    @property  # type: ignore
    @sdk_public_method
    def entity_names(self) -> List[str]:
        """
        Returns a list of entity names.
        """
        entity_protos = [self._fco_container.get_by_id(IdHelper.to_string(id)) for id in self._proto.entity_ids]
        return [entity.fco_metadata.name for entity in entity_protos]

    @property  # type: ignore
    @sdk_public_method
    def features(self) -> List[str]:
        """
        Returns the names of the (output) features.
        """
        join_keys = self.join_keys
        timestamp_key = self.timestamp_key
        return [
            col_name
            for col_name in self._view_schema.column_names()
            if col_name not in join_keys and col_name != timestamp_key
        ]

    @property  # type: ignore
    @sdk_public_method
    def url(self) -> str:
        """
        Returns a link to the Tecton Web UI.
        """
        return self._proto.web_url

    @sdk_public_method
    def summary(self) -> Displayable:
        """
        Returns various information about this feature definition, including the most critical metadata such
        as the name, owner, features, etc.
        """
        request = GetFeatureViewSummaryRequest()
        request.fco_locator.id.CopyFrom(self._id_proto)
        request.fco_locator.workspace = self.workspace

        response = metadata_service.instance().GetFeatureViewSummary(request)

        def value_formatter(key, value):
            if key == "featureStartTimeSeconds":
                t = datetime.fromtimestamp(int(value))
                return t.astimezone(timezone.utc).strftime("%Y-%m-%d %H:%M:%S %Z")
            return value

        return Displayable.from_fco_summary(response.fco_summary, value_formatter=value_formatter)

    def _construct_feature_set_config(self) -> FeatureSetConfig:
        feature_set_config = FeatureSetConfig()
        feature_set_config._add(FeatureDefinitionWrapper(self._proto, self._fco_container))
        # adding dependent feature views for odfv
        if self._proto.HasField("on_demand_feature_view"):
            inputs = find_dependent_feature_set_items(
                self._fco_container,
                self._proto.pipeline.root,
                visited_inputs={},
                fv_id=self.id,
                workspace_name=self.workspace,
            )
            feature_set_config._definitions_and_configs = feature_set_config._definitions_and_configs + inputs
        return feature_set_config

    # method overrided by OnDemandFeatureView class
    def _should_infer_timestamp_of_spine(
        self, timestamp_key: Optional[str], spine: Optional[Union[pysparkDF, pd.DataFrame]]
    ):
        return spine is not None and timestamp_key is None

    def _point_in_time_join(self, spine: pysparkDF, timestamp_key: Optional[str], from_source: bool) -> DataFrame:
        is_live_workspace = utils.is_live_workspace(self.workspace)
        if not from_source and not is_live_workspace and not self._proto.HasField("on_demand_feature_view"):
            raise errors.FD_GET_MATERIALIZED_FEATURES_FROM_DEVELOPMENT_WORKSPACE(self.name, self.workspace)

        if (
            not from_source
            and not self._proto.HasField("on_demand_feature_view")
            and not self._proto.materialization_enabled
        ):
            raise errors.FD_GET_FEATURES_MATERIALIZATION_DISABLED(self.name)

        feature_set_config = self._construct_feature_set_config()
        if self._proto.HasField("on_demand_feature_view"):
            utils.validate_spine_dataframe(spine, timestamp_key, list(self._request_context.arg_to_schema.keys()))
            # check for odfv with feature table input
            if any(fv.is_feature_table for fv in feature_set_config.feature_definitions):
                if not is_live_workspace:
                    raise errors.FV_WITH_FT_DEVELOPMENT_WORKSPACE
                elif from_source:
                    raise errors.FROM_SOURCE_WITH_FT
        else:
            utils.validate_spine_dataframe(spine, timestamp_key)

        spark = TectonContext.get_instance()._spark
        df = data_frame_helper.get_features_for_spine(
            spark, spine, feature_set_config, timestamp_key=timestamp_key, from_source=from_source
        )

        df = utils.filter_internal_columns(df)
        return DataFrame._create(df)

    def _get_historical_features(
        self,
        spine: Optional[Union[pysparkDF, pd.DataFrame, DataFrame]],
        timestamp_key: Optional[str],
        start_time: Optional[Union[pendulum.DateTime, datetime]],
        end_time: Optional[Union[pendulum.DateTime, datetime]],
        entities: Optional[Union[pysparkDF, pd.DataFrame, DataFrame]],
        from_source: bool,
        save: bool,
        save_as: Optional[str],
    ) -> DataFrame:
        has_point_in_time_join_params = spine is not None
        has_get_features_params = start_time is not None or end_time is not None or entities is not None

        if has_point_in_time_join_params:
            if has_get_features_params:
                raise errors.GET_HISTORICAL_FEATURES_WRONG_PARAMS(
                    ["start_time", "end_time", "entities"], "the spine parameter is provided"
                )
            if isinstance(spine, pd.DataFrame):
                spark = TectonContext.get_instance()._spark
                spine = spark.createDataFrame(spine)
            elif isinstance(spine, DataFrame):
                spine = spine.to_spark()
            if self._should_infer_timestamp_of_spine(timestamp_key, spine):
                timestamp_key = utils.infer_timestamp(spine)
            df = self._point_in_time_join(spine, timestamp_key, from_source)
        else:
            if timestamp_key is not None:
                raise errors.GET_HISTORICAL_FEATURES_WRONG_PARAMS(
                    ["timestamp_key"], "the spine parameter is not provided"
                )
            fd = FeatureDefinitionWrapper(self._proto, self._fco_container)
            df = get_features(
                fd,
                start_time=start_time,
                end_time=end_time,
                entities=entities,
                from_source=from_source,
            )

        if save or save_as is not None:
            return Dataset._create(
                df=df,
                save_as=save_as,
                workspace=self.workspace,
                feature_definition_id=self.id,
                spine=spine,
                timestamp_key=timestamp_key,
            )
        return df

    def _deletion_status(self, verbose=False, limit=1000, sort_columns=None, errors_only=False):
        materialization_attempts = self._get_materialization_status().materialization_attempts
        deletion_attempts = [
            attempt
            for attempt in materialization_attempts
            if attempt.data_source_type == DataSourceType.DATA_SOURCE_TYPE_DELETION
        ]
        column_names, materialization_status_rows = utils.format_materialization_attempts(
            deletion_attempts, verbose, limit, sort_columns, errors_only
        )

        return self._create_materialization_table(column_names, materialization_status_rows)

    def _materialization_status(self, verbose=False, limit=1000, sort_columns=None, errors_only=False):
        materialization_attempts = self._get_materialization_status().materialization_attempts
        column_names, materialization_status_rows = utils.format_materialization_attempts(
            materialization_attempts, verbose, limit, sort_columns, errors_only
        )

        return self._create_materialization_table(column_names, materialization_status_rows)

    def _create_materialization_table(self, column_names, materialization_status_rows):
        print("All the displayed times are in UTC time zone")

        # Setting `max_width=0` creates a table with an unlimited width.
        table = Displayable.from_items(headings=column_names, items=materialization_status_rows, max_width=0)
        # Align columns in the middle horizontally
        table._text_table.set_cols_align(["c" for _ in range(len(column_names))])

        return table

    def _get_materialization_status(self) -> MaterializationStatus:
        """
        Returns MaterializationStatus proto for the FeatureView.
        """
        request = GetMaterializationStatusRequest()
        request.feature_package_id.CopyFrom(self._id_proto)

        response = metadata_service.instance().GetMaterializationStatus(request)
        return response.materialization_status

    def _delete_keys(
        self,
        keys: Union[pysparkDF, pd.DataFrame],
        online: bool = True,
        offline: bool = True,
    ) -> None:
        if not offline and not online:
            raise errors.NO_STORE_SELECTED
        fd = FeatureDefinitionWrapper(self._proto, self._fco_container)
        if offline and any([x.offline_enabled for x in self._proto.materialization_state_transitions]):
            if not fd.offline_store_config.HasField("delta"):
                raise errors.OFFLINE_STORE_NOT_SUPPORTED
        if online and any([x.online_enabled for x in self._proto.materialization_state_transitions]):
            if fd.fv.materialization_params.HasField("online_store_params") and not fd.online_store_params.HasField(
                "dynamo"
            ):
                raise errors.ONLINE_STORE_NOT_SUPPORTED
        if isinstance(keys, pd.DataFrame):
            if len(keys) == 0:
                raise errors.EMPTY_ARGUMENT("join_keys")
            if len(keys.columns[keys.columns.duplicated()]):
                raise errors.DUPLICATED_COLS_IN_KEYS(", ".join(list(keys.columns)))
            spark_df = self._convert_pandas_to_spark_df(keys)
        elif isinstance(keys, pysparkDF):
            spark_df = keys
        else:
            raise errors.INVALID_KEY_TYPE(type(keys))
        utils.validate_entity_deletion_keys_dataframe(df=spark_df, join_keys=fd.join_keys, view_schema=fd.view_schema)
        is_live_workspace = utils.is_live_workspace(self.workspace)
        if not is_live_workspace:
            raise errors.UNSUPPORTED_OPERATION_IN_DEVELOPMENT_WORKSPACE("delete_keys")

        if online and all([not x.online_enabled for x in self._proto.materialization_state_transitions]):
            print("Online materialization was never enabled. No data to be deleted in online store.")
            online = False

        if offline and all([not x.offline_enabled for x in self._proto.materialization_state_transitions]):
            print("Offline materialization was never enabled. No data to be deleted in offline store.")
            offline = False

        if online or offline:
            self._send_delete_keys_request(keys, online, offline)

    def _send_delete_keys_request(self, keys: Union[pysparkDF, pd.DataFrame], online: bool, offline: bool):
        def serialize_fn(x):
            ret = FeatureValueList()
            for item in x:
                if isinstance(item, int):
                    ret.feature_values.add().int64_value = item
                elif isinstance(item, str):
                    ret.feature_values.add().string_value = item
                elif item is None:
                    ret.feature_values.add().null_value.CopyFrom(NullValue())
                else:
                    raise Exception(f"Unknown type: {type(item)}")
            return b64encode(ret.SerializeToString()).decode()

        info_request = GetDeleteEntitiesInfoRequest()
        info_request.feature_definition_id.CopyFrom(self._id_proto)
        info_response = metadata_service.instance().GetDeleteEntitiesInfo(info_request)
        s3_path = info_response.df_path
        online_join_keys_path = s3_path + "/online"
        offline_join_keys_path = s3_path + "/offline"
        deletion_request = DeleteEntitiesRequest()
        deletion_request.fco_locator.id.CopyFrom(self._id_proto)
        deletion_request.fco_locator.workspace = self.workspace
        deletion_request.online = online
        deletion_request.offline = offline
        deletion_request.online_join_keys_path = online_join_keys_path
        deletion_request.offline_join_keys_path = offline_join_keys_path
        if online:
            spark = TectonContext.get_instance()._spark
            if isinstance(keys, pd.DataFrame):
                spark_keys_df = spark.createDataFrame(keys)
            else:
                spark_keys_df = keys
            spark_keys_df = spark_keys_df.distinct()
            serialize = udf(serialize_fn, StringType())
            join_key_df = spark_keys_df.withColumn("join_keys_array", array(*self.join_keys)).select(
                serialize("join_keys_array")
            )

            # coalesce(1) causes it to write to 1 file, but the jvm code
            # is actually robust to multiple files here
            join_key_df.coalesce(1).write.csv(online_join_keys_path)

        if offline:
            # We write in the native format and avoid converting Pandas <-> Spark due to partially incompatible
            # type system, in specifically missing Int in Pandas
            if isinstance(keys, pysparkDF):
                keys.write.parquet(offline_join_keys_path)
            else:
                upload_url = info_response.signed_url_for_df_upload
                self._check_types_and_upload_df_pandas(upload_url, offline_join_keys_path, keys)

        metadata_service.instance().DeleteEntities(deletion_request)
        print(
            "A deletion job has been created. You can track the status of the job in the Web UI under Materialization section or with deletion_status(). The deletion jobs have a type 'Deletion'."
        )

    def _convert_pandas_to_spark_df(self, df: pd.DataFrame):
        tc = TectonContext.get_instance()
        spark = tc._spark
        spark_df = spark.createDataFrame(df)

        converted_schema = self._convert_ingest_schema(spark_df.schema)

        if converted_schema != spark_df.schema:
            spark_df = spark.createDataFrame(df, schema=converted_schema)

        return spark_df

    def _check_types_and_upload_df_pandas(self, upload_url: str, df_path: str, df: pd.DataFrame):
        spark_df = self._convert_pandas_to_spark_df(df)

        spark_df.write.parquet(df_path)
        # Theoretically a faster way to upload native pandas DF, compared to converting to Spark DF first.
        # Currently disabled due to dependency version incompatibility between numpy==1.21.0 in DB notebooks
        # with pyarrow==0.13.0 that we pin. Can't upgrade pyarrow due to incompatibility with
        # Spark 2, see https://tecton-ai.phacility.com/D8180#141519.
        # self._upload_df_pandas(upload_url, df)

    def _convert_ingest_schema(self, ingest_schema: StructType) -> StructType:
        """
        The Pandas to Spark dataframe conversion implicitly derives the Spark schema. We handle converting/correcting
        for some type conversions where the derived schema and the feature table schema do not match.
        """
        ft_columns = column_name_spark_data_types(self._view_schema)
        ingest_columns = column_name_spark_data_types(schema_from_spark(ingest_schema))

        converted_ingest_schema = StructType()
        int_converted_columns = []

        for col_name, col_type in ingest_columns:
            if col_type == LongType() and (col_name, IntegerType()) in ft_columns:
                int_converted_columns.append(col_name)
                converted_ingest_schema.add(col_name, IntegerType())
            elif col_type == ArrayType(DoubleType()) and (col_name, ArrayType(FloatType())) in ft_columns:
                converted_ingest_schema.add(col_name, ArrayType(FloatType()))
            else:
                converted_ingest_schema.add(col_name, col_type)

        if int_converted_columns:
            logger.warning(
                f"Tecton is casting field(s) {', '.join(int_converted_columns)} to type Integer (was type Long). To remove this warning, use a Long type in the schema."
            )

        return converted_ingest_schema



import functools
from dataclasses import dataclass
from dataclasses import field
from inspect import signature
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

from google.protobuf.empty_pb2 import Empty
from pyspark.sql import DataFrame
from typeguard import typechecked

from tecton._internals import errors
from tecton._internals.fco import Fco
from tecton.basic_info import prepare_basic_info
from tecton.cli.common import get_fco_source_info
from tecton.transformations.const import Constant
from tecton_proto.args.new_transformation_pb2 import NewTransformationArgs as TransformationArgs
from tecton_proto.args.new_transformation_pb2 import TransformationMode
from tecton_proto.args.pipeline_pb2 import Input
from tecton_proto.args.pipeline_pb2 import MaterializationContextNode
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_proto.args.repo_metadata_pb2 import SourceInfo
from tecton_proto.common.id_pb2 import Id
from tecton_spark import function_serialization as func_ser
from tecton_spark.id_helper import IdHelper
from tecton_spark.materialization_context import materialization_context
from tecton_spark.materialization_context import UnboundMaterializationContext

SPARK_SQL_MODE = "spark_sql"
SNOWFLAKE_SQL_MODE = "snowflake_sql"
PYSPARK_MODE = "pyspark"
PANDAS_MODE = "pandas"
PYTHON_MODE = "python"

_GLOBAL_TRANSFORMATIONS_LIST = []


@dataclass
class Transformation(Fco):
    """
    (Tecton Object) Transformation class.
    """

    name: str
    mode: str
    user_function: Callable[..., Union[str, DataFrame]]
    description: str
    owner: str
    family: str
    tags: Dict[str, str]
    _source_info: SourceInfo = field(init=False, repr=False)
    _args: TransformationArgs = field(init=False, repr=False)
    _ARGUMENT_TYPE = Union[PipelineNode, Constant, UnboundMaterializationContext]

    def _docstring(self):
        return None

    @property
    def transformer(self):
        return func_ser.from_proto(self._args.user_function)

    def _is_builtin(self):
        return False

    def __post_init__(self):
        self._source_info = get_fco_source_info()

        self._args = TransformationArgs()
        self._args.transformation_id.CopyFrom(IdHelper.from_string(IdHelper.generate_string_id()))
        self._args.info.CopyFrom(
            prepare_basic_info(
                name=self.name, description=self.description, owner=self.owner, family=self.family, tags=self.tags
            )
        )
        if self.mode == SPARK_SQL_MODE:
            transform_mode = TransformationMode.TRANSFORMATION_MODE_SPARK_SQL
        elif self.mode == SNOWFLAKE_SQL_MODE:
            transform_mode = TransformationMode.TRANSFORMATION_MODE_SNOWFLAKE_SQL
        elif self.mode == PYSPARK_MODE:
            transform_mode = TransformationMode.TRANSFORMATION_MODE_PYSPARK
        elif self.mode == PANDAS_MODE:
            transform_mode = TransformationMode.TRANSFORMATION_MODE_PANDAS
        elif self.mode == PYTHON_MODE:
            transform_mode = TransformationMode.TRANSFORMATION_MODE_PYTHON
        else:
            raise errors.InvalidTransformationMode(self.name, self.mode, [SPARK_SQL_MODE, PYSPARK_MODE])
        if self._docstring() is not None:
            self._args.docstring = self._docstring()
        self._args.transformation_mode = transform_mode
        self._args.user_function.CopyFrom(func_ser.to_proto(self.user_function))
        self._args.is_builtin = self._is_builtin()
        if not self._is_builtin():
            Fco._register(self)
        _GLOBAL_TRANSFORMATIONS_LIST.append(self)

    def _args_proto(self) -> TransformationArgs:
        return self._args

    def __call__(self, *args, **kwargs) -> PipelineNode:
        wrapper = PipelineNode()
        node = wrapper.transformation_node
        node.transformation_id.CopyFrom(self._args.transformation_id)

        try:
            bound_user_function = signature(self.user_function).bind(*args, **kwargs)
        except TypeError as e:
            raise TypeError(f"while binding inputs to function {self.name}, TypeError: {e}")

        materialization_context_count = 0

        for i, arg in enumerate(args):
            input_ = Input()
            input_.arg_index = i
            input_.node.CopyFrom(self._value_to_node(arg))
            node.inputs.append(input_)
            if isinstance(arg, UnboundMaterializationContext):
                materialization_context_count += 1
        for name, arg in kwargs.items():
            input_ = Input()
            input_.arg_name = name
            input_.node.CopyFrom(self._value_to_node(arg))
            node.inputs.append(input_)
            if isinstance(arg, UnboundMaterializationContext):
                materialization_context_count += 1

        for param in signature(self.user_function).parameters.values():
            if isinstance(param.default, UnboundMaterializationContext):
                if param.name in bound_user_function.arguments:
                    # the user passed in context explicitly, so no need to double register
                    continue
                input_ = Input()
                input_.arg_name = param.name
                input_.node.CopyFrom(self._value_to_node(param.default))
                node.inputs.append(input_)
                if isinstance(param.default, UnboundMaterializationContext):
                    materialization_context_count += 1
            elif param.default is materialization_context:
                raise Exception(
                    "It seems you passed in tecton.materialization_context. Did you mean tecton.materialization_context()?"
                )

        if materialization_context_count > 1:
            raise Exception(f"Only 1 materialization_context can be passed into transformation {self.name}")

        return wrapper

    def _value_to_node(self, arg: _ARGUMENT_TYPE) -> PipelineNode:
        if isinstance(arg, PipelineNode):
            return arg
        elif isinstance(arg, Constant):
            wrapper = PipelineNode()
            node = wrapper.constant_node
            if arg.value is None:
                node.null_const.CopyFrom(Empty())
            elif arg.value_type == str:
                node.string_const = arg.value
            elif arg.value_type == int:
                node.int_const = repr(arg.value)
            elif arg.value_type == float:
                node.float_const = repr(arg.value)
            elif arg.value_type == bool:
                node.bool_const = arg.value
            return wrapper
        elif isinstance(arg, UnboundMaterializationContext):
            wrapper = PipelineNode()
            wrapper.materialization_context_node.CopyFrom(MaterializationContextNode())
            assert wrapper.HasField("materialization_context_node")
            return wrapper
        else:
            raise errors.InvalidTransformInvocation(self.name, arg)

    def _id(self) -> Id:
        return self._args.transformation_id

    def __hash__(self):
        return hash(self.name)


@typechecked
def transformation(
    mode: str,
    name_override: Optional[str] = None,
    description: Optional[str] = None,
    owner: Optional[str] = None,
    family: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
):
    """
    Declares a Transformation that wraps a user function. Transformations are assembled in a pipeline function of a Feature View.

    :param mode: The mode for this transformation must be one of "spark_sql", "pyspark", "pandas" or "python".
    :param name_override: Unique, human friendly name override that identifies the Transformation.
    :param description: Human readable description.
    :param owner: Owner name (typically the email of the primary maintainer).
    :param family: Family of this Transformation, used to group Tecton Objects.
    :param tags: Tags associated with this Tecton Object (key-value pairs of arbitrary metadata).
    :return: A wrapped transformation

    Examples of Spark SQL, PySpark, Pandas, and Python transformation declarations:

        .. code-block:: python

            from tecton import transformation
            import pyspark.sql.DataFrame
            import pandas as pd

            # Create a Spark SQL transformation.
            @transformation(mode="spark_sql",
                            description="Create new column by splitting the string in an existing column")
            def str_split(input_data, column_to_split, new_column_name, delimiter):
                return f'''
                    SELECT
                        *,
                        split({column_to_split}, {delimiter}) AS {new_column_name}
                    FROM {input_data}'''

            # Create a PySpark transformation.
            @transformation(mode="pyspark",
                            description="Add a new column 'user_has_good_credit' if score is > 670")
            def user_has_good_credit_transformation(credit_scores):
                from pyspark.sql import functions as F

                (df = credit_scores.withColumn("user_has_good_credit",
                    F.when(credit_scores["credit_score"] > 670, 1).otherwise(0))
                return df.select("user_id", df["date"].alias("timestamp"), "user_has_good_credit") )

            # Create a Pandas transformation.
            @transformation(mode="pandas",
                            description="Whether the transaction amount is considered high (over $10000)")
            def transaction_amount_is_high(transaction_request):
                import pandas as pd

                df = pd.DataFrame()
                df['amount_is_high'] = (request['amount'] >= 10000).astype('int64')
                return df

            @transformation(mode="python",
                            description="Whether the transaction amount is considered high (over $10000)")
            # Create a Python transformation.
            def transaction_amount_is_high(transaction_request):

                result = {}
                result['transaction_amount_is_high'] = int(transaction_request['amount'] >= 10000)
                return result
    """

    def decorator(user_function):
        transform_name = name_override or user_function.__name__
        transform = Transformation(transform_name, mode, user_function, description, owner, family, tags)
        functools.update_wrapper(wrapper=transform, wrapped=user_function)

        return transform

    return decorator



import base64
import json
from typing import Optional
from typing import Union

import pandas
import pyspark
from google.protobuf.json_format import MessageToJson
from pyspark.sql.types import StructType

from tecton import conf
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import sdk_public_method
from tecton.interactive.data_frame import DataFrame
from tecton.tecton_context import TectonContext
from tecton_proto.data.saved_feature_data_frame_pb2 import SavedFeatureDataFrame
from tecton_proto.data.saved_feature_data_frame_pb2 import SavedFeatureDataFrameType
from tecton_proto.metadataservice.metadata_service_pb2 import ArchiveSavedFeatureDataFrameRequest
from tecton_proto.metadataservice.metadata_service_pb2 import CreateSavedFeatureDataFrameRequest
from tecton_proto.metadataservice.metadata_service_pb2 import GetSavedFeatureDataFrameRequest
from tecton_spark.id_helper import IdHelper
from tecton_spark.logger import get_logger
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper

logger = get_logger("FeatureService")


class Dataset(DataFrame):
    """
    Dataset class.

    Persisted data consisting of entity & request keys, timestamps, and calculated features. Datasets are
    associated with either a :class:`FeatureService` or :class:`FeatureView`.

    There are 2 types of Datasets: Saved and Logged.

    Saved Datasets are generated manually when calling :meth:`tecton.get_historical_features` by setting the ``save``
    parameter.

    Logged Datasets are generated automatically when declaring a :class:`FeatureService` with :class:`tecton.LoggingConfig`,
    and the data is continuously added to it when requesting online data from the FeatureService.

    To get an existing Dataset, call :py:meth:`tecton.get_dataset`.
    """

    _proto: SavedFeatureDataFrame = None

    def __init__(self, proto, spark_df, pandas_df):
        super().__init__(spark_df, pandas_df, None)
        self._proto = proto
        self.is_archived = proto.info.is_archived

    @classmethod
    def _from_proto(cls, proto):
        return cls(proto, None, None)

    @sdk_public_method
    def to_spark(self) -> pyspark.sql.DataFrame:
        """Converts the Dataset to a Spark DataFrame and returns it."""
        self._try_fetch_spark_df()
        return super().to_spark()

    @sdk_public_method
    def to_pandas(self) -> pandas.DataFrame:
        """Converts the Dataset to a Pandas DataFrame and returns it."""
        self._try_fetch_spark_df()
        return super().to_pandas()

    # Creates and returns an empty Spark dataframe & pandas dataframe with desired schema
    def _create_empty_dfs(self):
        schema_json = json.loads(MessageToJson(self._proto.schema))
        fields = []
        for field in schema_json["fields"]:
            fields.append(json.loads(field["structfieldJson"]))

        schema = StructType.fromJson({"fields": fields})

        spark = TectonContext.get_instance()._get_spark()

        spark_df = spark.createDataFrame(spark.sparkContext.emptyRDD(), schema)
        pandas_df = pandas.DataFrame(columns=[field.name for field in schema.fields])
        return spark_df, pandas_df

    # Tries fetching self._spark_df. As long as the underlying data exists,
    # it's expected to succeed. However, in certain cases self._spark_df may stay None.
    # For example, if this is a logged dataset and there are not feature requests logged
    # yet, self._spark_df will stay None after the execution of this method.
    def _try_fetch_spark_df(self):
        if self._spark_df is not None:
            return
        spark = TectonContext.get_instance()._get_spark()
        try:
            if self._type == SavedFeatureDataFrameType.LOGGED:
                # Logged datasets are in Avro format
                self._spark_df = spark.read.format("avro").load(self._path)
                self._spark_df = _convert_logged_df_schema(self._spark_df)
            else:
                self._spark_df = spark.read.parquet(self._path)
        except pyspark.sql.utils.AnalysisException as e:
            # If the path doesn't exist in S3, there is no data
            # This can happen for logged features when there is no logs yet,
            # so we don't want to throw an error in this case
            if "Path does not exist" in e.desc:
                self._spark_df, self._pandas_df = self._create_empty_dfs()
            else:
                raise e

    @sdk_public_method
    def summary(self) -> Displayable:
        """
        Print out a summary of this class's attributes.
        """
        return Displayable.from_items(headings=["Property", "Value"], items=self._summary_items())

    def _summary_items(self):
        items = [
            ("Name", self.name),
            ("Id", IdHelper.to_string(self._proto.saved_feature_dataframe_id)),
            ("Created At", self._proto.info.created_at.ToJsonString()),
            ("Workspace", self._proto.info.workspace or "prod"),
            ("Tecton Log Commit Id", self._proto.state_update_entry_commit_id),
            ("Type", "Logged" if self._type == SavedFeatureDataFrameType.LOGGED else "Saved"),
        ]
        items.append(self._get_source())
        if len(self._proto.join_key_column_names) > 0:
            items.append(("Join & Request Keys", ", ".join(self._proto.join_key_column_names)))
        if self._proto.HasField("timestamp_column_name"):
            items.append(("Timestamp Key", self._proto.timestamp_column_name))
        return items

    def _get_source(self):
        if self._proto.HasField("feature_package_name"):
            return ("Source FeatureView", self._proto.feature_package_name)
        elif self._proto.HasField("feature_service_name"):
            return ("Source FeatureService", self._proto.feature_service_name)
        else:
            # should be unreachable
            assert False, "Neither feature_package_name nor feature_service_name set in the proto"

    def _delete(self):
        """
        Delete this Dataset. Note that this deletes the underlying data as well as removing the Dataset object from
        Tecton.
        """
        self.is_archived = True
        request = ArchiveSavedFeatureDataFrameRequest()
        request.saved_feature_dataframe_id.CopyFrom(IdHelper.from_string(self._id))
        metadata_service.instance().ArchiveSavedFeatureDataFrame(request)
        logger.info(f"Dataset {self.name} deleted")

    @sdk_public_method
    def get_spine_dataframe(self) -> DataFrame:
        """
        Get a :py:class:`tecton.DataFrame` containing the spine.
        """
        if not (self._proto.join_key_column_names and self._proto.timestamp_column_name):
            raise errors.DATASET_SPINE_COLUMNS_NOT_SET
        self._try_fetch_spark_df()
        spine_spark_df = self._spark_df.select(
            self._proto.join_key_column_names[:] + [self._proto.timestamp_column_name]
        )
        return DataFrame(spine_spark_df, None)

    @property
    def name(self):
        """
        Dataset name
        """
        return self._proto.info.name

    @property
    def _id(self):
        return IdHelper.to_string(self._proto.saved_feature_dataframe_id)

    @property
    def _path(self):
        return self._proto.dataframe_location

    @property
    def _feature_service_id(self):
        return IdHelper.to_string(self._proto.feature_service_id)

    @property
    def _type(self):
        return self._proto.type

    @classmethod
    def _create(
        cls,
        df: DataFrame,
        save_as: Optional[str] = None,
        workspace: Optional[str] = None,
        feature_definition_id: Optional[str] = None,
        feature_service_id: Optional[str] = None,
        spine: Optional[Union[pyspark.sql.dataframe.DataFrame, pandas.DataFrame]] = None,
        timestamp_key: Optional[str] = None,
    ) -> DataFrame:
        assert (
            feature_definition_id or feature_service_id
        ), "Either feature_definition_id or feature_service_id must be provided"

        create_request = CreateSavedFeatureDataFrameRequest()
        if save_as:
            create_request.name = save_as
        create_request.workspace = workspace
        if feature_definition_id:
            create_request.feature_package_id.CopyFrom(IdHelper.from_string(feature_definition_id))
        if feature_service_id:
            create_request.feature_service_id.CopyFrom(IdHelper.from_string(feature_service_id))
        if spine is not None:
            # note that spine.columns works for both pandas and spark dfs
            spine_cols = [s for s in spine.columns if s != timestamp_key]
            create_request.join_key_column_names.extend(spine_cols)
            if timestamp_key is not None:  # timestamp_key can be none for odfvs
                create_request.timestamp_column_name = timestamp_key
        create_request.schema.CopyFrom(SparkSchemaWrapper.from_spark_schema(df.to_spark().schema))
        response = metadata_service.instance().CreateSavedFeatureDataFrame(create_request)
        proto = response.saved_feature_dataframe
        logger.info(f"Saved Dataset with name {proto.info.name}")
        logger.debug(f"Path is {proto.dataframe_location}")

        df.to_spark().write.save(path=proto.dataframe_location, format="parquet", mode="errorifexists")
        return cls(proto, df._spark_df, df._pandas_df)

    def __repr__(self):
        source_type, source_value = self._get_source()
        source_str = f"{source_type}='{source_value}'"
        return (
            f"{type(self).__name__}(name='{self.name}', "
            + f"{source_str}, created_at='{self._proto.info.created_at.ToJsonString()}')"
        )


@sdk_public_method
def get_dataset(name: str, workspace_name: Optional[str] = None) -> Dataset:
    """
    Returns a Dataset with a given name.

    :param name: Dataset name.
    :return: An instance of :class:`tecton.interactive.Dataset`
    """
    request = GetSavedFeatureDataFrameRequest()
    request.saved_feature_dataframe_name = name
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetSavedFeatureDataFrame(request)

    return Dataset._from_proto(response.saved_feature_dataframe)


def get_dataset_by_id(_id: str, workspace_name: Optional[str] = None) -> Dataset:
    request = GetSavedFeatureDataFrameRequest()
    request.saved_feature_dataframe_id.CopyFrom(IdHelper.from_string(_id))
    request.workspace = workspace_name or conf.get_or_none("TECTON_WORKSPACE")
    response = metadata_service.instance().GetSavedFeatureDataFrame(request)
    return Dataset._from_proto(response.saved_feature_dataframe)


def _convert_logged_df_schema(spark_df):
    if spark_df is None:
        return
    # Note: _partition column is not used right now, but in future
    # it can be used to optimize time-range access of this dataframe
    spark_df = spark_df.drop("_partition")
    # Note: the rest of the column names are base16 encoded due to strict
    # Avro column name validation (only [_a-zA-Z0-9] allowed). The encoding
    # happens here:
    for column in spark_df.columns:
        new_column = base64.b16decode(column[1:], casefold=True).decode()
        spark_df = spark_df.withColumnRenamed(column, new_column)
    return spark_df



from collections.abc import Callable
from typing import Optional

from pyspark.sql.types import StructType

from tecton.data_sources.base_data_source_config import BaseBatchDSConfig
from tecton_proto.args.data_source_pb2 import FileDataSourceArgs
from tecton_proto.args.virtual_data_source_pb2 import VirtualDataSourceArgs
from tecton_spark import function_serialization
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper


class FileDSConfig(BaseBatchDSConfig):
    """
    Configuration used to reference a file or directory (S3, etc.)

    The FileDSConfig class is used to create a reference to a file or directory of files in S3,
    HDFS, or DBFS.

    The schema of the data source is inferred from the underlying file(s). It can also be modified using the
    ``raw_batch_translator`` parameter.

    This class is used as an input to a :class:`BatchDataSource`'s parameter ``batch_ds_config``. This class is not
    a Tecton Object: it is a grouping of parameters. Declaring this class alone will not register a data source.
    Instead, declare a part of ``BatchDataSource`` that takes this configuration class instance as a parameter.
    """

    def __init__(
        self,
        uri: str,
        file_format: str,
        convert_to_glue_format=False,
        timestamp_column_name: Optional[str] = None,
        timestamp_format: Optional[str] = None,
        raw_batch_translator: Optional[Callable] = None,
        schema_uri: Optional[str] = None,
        schema_override: Optional[StructType] = None,
    ):
        """
        Instantiates a new FileDSConfig.

        :param uri: S3 or HDFS path to file(s).
        :param file_format: File format. "json", "parquet", or "csv"
        :param convert_to_glue_format: Converts all schema column names to lowercase.
        :param timestamp_column_name: Name of timestamp column.
        :param timestamp_format: (Optional) Format of string-encoded timestamp column (e.g. "yyyy-MM-dd'T'hh:mm:ss.SSS'Z'").
                                 If the timestamp string cannot be parsed with this format, Tecton will fallback and attempt to
                                 use the default timestamp parser.
        :param raw_batch_translator: Python user defined function f(DataFrame) -> DataFrame that takes in raw
                                     Pyspark data source DataFrame and translates it to the DataFrame to be
                                     consumed by the Feature View. See an example of
                                     raw_batch_translator in the `User Guide`_.
        :param schema_uri: (Optional) A file or subpath of "uri" that can be used for fast schema inference.
                           This is useful for speeding up plan computation for highly partitioned data sources containing many files.
        :param schema_override: (Optional) a pyspark.sql.types.StructType object that will be used as the schema when
                                reading from the file. If omitted, the schema will be inferred automatically.

        :return: A FileDSConfig class instance.

        .. _User Guide: https://docs.tecton.ai/v2/overviews/framework/data_sources.html

        Example of a FileDSConfig declaration:

        .. code-block:: python

            from tecton import FileDSConfig, BatchDataSource

            def convert_temperature(df):
                from pyspark.sql.functions import udf,col
                from pyspark.sql.types import DoubleType

                # Convert the incoming PySpark DataFrame temperature Celsius to Fahrenheit
                udf_convert = udf(lambda x: x * 1.8 + 32.0, DoubleType())
                converted_df = df.withColumn("Fahrenheit", udf_convert(col("Temperature"))).drop("Temperature")
                return converted_df

            # declare a FileDSConfig, which can be used as a parameter to a `BatchDataSource`
            ad_impressions_file_ds = FileDSConfig(uri="s3://tecton.ai.public/data/ad_impressions_sample.parquet",
                                                file_format="parquet",
                                                timestamp_column_name="timestamp",
                                                raw_batch_translator=convert_temperature)

            # This FileDSConfig can then be included as an parameter a BatchDataSource declaration.
            # For example,
            ad_impressions_batch = BatchDataSource(name="ad_impressions_batch",
                                                   batch_ds_config=ad_impressions_file_ds)

        """
        self._args = prepare_file_ds_args(
            uri=uri,
            file_format=file_format,
            convert_to_glue_format=convert_to_glue_format,
            timestamp_column_name=timestamp_column_name,
            timestamp_format=timestamp_format,
            raw_batch_translator=raw_batch_translator,
            schema_uri=schema_uri,
            schema_override=schema_override,
        )

    def _merge_batch_args(self, data_source_args: VirtualDataSourceArgs):
        data_source_args.file_ds_config.CopyFrom(self._args)


def prepare_file_ds_args(
    *,
    uri: str,
    file_format: str,
    convert_to_glue_format: bool,
    timestamp_column_name: Optional[str],
    timestamp_format: Optional[str],
    raw_batch_translator: Optional[Callable],
    schema_uri: Optional[str],
    schema_override: Optional[StructType],
) -> FileDataSourceArgs:
    args = FileDataSourceArgs()
    args.uri = uri
    args.file_format = file_format
    args.convert_to_glue_format = convert_to_glue_format
    if schema_uri is not None:
        args.schema_uri = schema_uri
    if raw_batch_translator is not None:
        args.raw_batch_translator.CopyFrom(function_serialization.to_proto(raw_batch_translator))
    if timestamp_column_name:
        args.timestamp_column_name = timestamp_column_name
    if timestamp_format:
        args.timestamp_format = timestamp_format
    if schema_override:
        args.schema_override.CopyFrom(SparkSchemaWrapper(schema_override).to_proto())

    return args



import click

from tecton.cli.command import tecton_config_option
from tecton.cli.command import TectonGroup
from tecton_spark import conf


# TODO(snowflake): Unhide
@click.command(cls=TectonGroup, hidden=True)
def snowflake():
    """Snowflake-related commands."""


@snowflake.command(requires_auth=False)
@tecton_config_option(
    "SNOWFLAKE_USER", param_decls=["--user"], help="The Snowflake user to use for interactive queries."
)
@tecton_config_option(
    "SNOWFLAKE_PASSWORD", param_decls=["--password"], hide_input=True, help="The password for the Snowflake user."
)
@tecton_config_option(
    "SNOWFLAKE_WAREHOUSE", param_decls=["--warehouse"], help="The Snowflake warehouse to use for interactive queries."
)
def configure(user, password, warehouse):
    conf.set("ALPHA_SNOWFLAKE_COMPUTE_ENABLED", "true")
    conf.set("SNOWFLAKE_USER", user)
    conf.set("SNOWFLAKE_PASSWORD", password)
    conf.set("SNOWFLAKE_WAREHOUSE", warehouse)
    conf.save_tecton_configs()



import tempfile
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import pandas
import pendulum
from pyspark.sql import DataFrame
from pyspark.sql.streaming import StreamingQuery

import tecton
from tecton._internals import errors
from tecton._internals.feature_views.aggregations import construct_full_tafv_df
from tecton.run_api_consts import AGGREGATION_LEVEL_DISABLED
from tecton.run_api_consts import AGGREGATION_LEVEL_FULL
from tecton.run_api_consts import AGGREGATION_LEVEL_PARTIAL
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME
from tecton.run_api_consts import SUPPORTED_AGGREGATION_LEVEL_VALUES
from tecton.tecton_context import TectonContext
from tecton_proto.args.feature_view_pb2 import BackfillConfigMode
from tecton_proto.data import feature_view_pb2
from tecton_spark import logger as logger_lib
from tecton_spark import materialization_plan
from tecton_spark import time_utils
from tecton_spark.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_spark.id_helper import IdHelper
from tecton_spark.partial_aggregations import construct_partial_time_aggregation_df
from tecton_spark.partial_aggregations import rename_partial_aggregate_columns
from tecton_spark.pipeline_helper import get_all_input_ds_id_map
from tecton_spark.pipeline_helper import get_all_input_keys
from tecton_spark.pipeline_helper import pipeline_to_dataframe
from tecton_spark.pipeline_helper import run_mock_odfv_pipeline
from tecton_spark.spark_schema_wrapper import SparkSchemaWrapper

logger = logger_lib.get_logger("InteractiveRunApi")


def _resolve_times(
    fd: FeatureDefinition,
    feature_start_time: Optional[Union[pendulum.DateTime, datetime]],
    feature_end_time: Optional[Union[pendulum.DateTime, datetime]],
    aggregation_level: str,
) -> Tuple[datetime, datetime, pendulum.Period]:
    schedule_interval = fd.min_scheduling_interval
    fv_proto = fd.feature_view_proto

    # Smart default feature_start_time and feature_end_time if unset.
    if feature_end_time is None:
        feature_end_time = pendulum.now()
    if feature_start_time is None:
        feature_start_time = feature_end_time - schedule_interval

    # If doing a "full" aggregation, move back the feature_start_time by the
    # duration to ensure we have enough data.
    if aggregation_level == AGGREGATION_LEVEL_FULL:
        max_aggregation = max((f.window.seconds for f in fd.trailing_time_window_aggregation.features))
        feature_start_time = feature_start_time - pendulum.duration(seconds=max_aggregation)

    if isinstance(feature_start_time, pendulum.DateTime):
        feature_start_time = datetime.utcfromtimestamp(feature_start_time.timestamp())
    if isinstance(feature_end_time, pendulum.DateTime):
        feature_end_time = datetime.utcfromtimestamp(feature_end_time.timestamp())

    feature_time_limits_aligned = _align_times(feature_start_time, feature_end_time, fv_proto, schedule_interval)

    _validate_feature_time_for_backfill_config(
        fv_proto, feature_start_time, feature_end_time, feature_time_limits_aligned
    )

    return feature_start_time, feature_end_time, feature_time_limits_aligned


def run_batch(
    fd: FeatureDefinition,
    feature_start_time: Optional[Union[pendulum.DateTime, datetime]],
    feature_end_time: Optional[Union[pendulum.DateTime, datetime]],
    mock_inputs: Dict[str, Union[pandas.DataFrame, DataFrame]],
    aggregate_tiles: Optional[bool] = None,
    aggregation_level: Optional[str] = None,
) -> "tecton.interactive.data_frame.DataFrame":
    fv_proto = fd.feature_view_proto
    spark = TectonContext.get_instance()._spark
    aggregation_level = validate_and_get_aggregation_level(fd, aggregate_tiles, aggregation_level)

    # Validate that mock_inputs' keys.
    input_ds_id_map = get_all_input_ds_id_map(fd.pipeline.root)
    _validate_batch_mock_inputs_keys(mock_inputs, fv_proto)

    feature_start_time, feature_end_time, feature_time_limits_aligned = _resolve_times(
        fd, feature_start_time, feature_end_time, aggregation_level
    )

    # Convert any Pandas dataFrame mock_inputs to Spark, validate schema columns.
    # TODO(raviphol): Consider refactor this under pipeline_helper._node_to_value
    data_sources = fd.data_sources
    for key in mock_inputs.keys():
        ds = _get_ds_by_id(data_sources, input_ds_id_map[key])
        spark_schema = _get_spark_schema(ds)

        if isinstance(mock_inputs[key], pandas.DataFrame):
            mock_inputs[key] = spark.createDataFrame(mock_inputs[key])

        _validate_input_dataframe_schema(input_name=key, dataframe=mock_inputs[key], spark_schema=spark_schema)

    # Execute Spark pipeline to get output DataFrame.
    materialized_spark_df = pipeline_to_dataframe(
        spark,
        pipeline=fd.pipeline,
        consume_streaming_data_sources=False,
        data_sources=data_sources,
        transformations=fd.transformations,
        feature_time_limits=feature_time_limits_aligned,
        schedule_interval=pendulum.Duration(seconds=fv_proto.materialization_params.schedule_interval.ToSeconds()),
        mock_inputs=mock_inputs,
    )

    if aggregation_level == "partial":
        # Aggregate the output rows into corresponding aggregate-tiles.
        materialized_spark_df = construct_partial_time_aggregation_df(
            df=materialized_spark_df,
            join_keys=fd.join_keys,
            time_aggregation=fd.trailing_time_window_aggregation,
            version=fv_proto.feature_store_format_version,
            window_start_column_name=DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME,
            window_end_column_name=DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME,
            convert_to_epoch=False,
        )
        # Intermediate-rollup output columns will be renamed to use the similar pattern as final aggregated columns.
        materialized_spark_df = rename_partial_aggregate_columns(
            df=materialized_spark_df,
            slide_interval_string=fd.get_aggregate_slide_interval_string,
            trailing_time_window_aggregation=fd.trailing_time_window_aggregation,
        )
    elif aggregation_level == AGGREGATION_LEVEL_FULL:
        # Aggregate the output rows into corresponding aggregate-tiles.
        materialized_spark_df = construct_partial_time_aggregation_df(
            df=materialized_spark_df,
            join_keys=fd.join_keys,
            time_aggregation=fd.trailing_time_window_aggregation,
            version=fv_proto.feature_store_format_version,
        )
        # Perform final rollup from aggregate tiles up to each result window.
        materialized_spark_df = construct_full_tafv_df(
            spark=spark,
            time_aggregation=fd.trailing_time_window_aggregation,
            join_keys=fd.join_keys,
            feature_store_format_version=fv_proto.feature_store_format_version,
            tile_interval=fd.get_tile_interval,
            all_partial_aggregations_df=materialized_spark_df,
            use_materialized_data=False,
        )
    else:
        # Filter output rows which are not within feature time range. This won't apply if aggregate_tiles is set,
        # because aggregated output rows won't correspond to a timestamp, but instead aggregated into the time window.
        materialized_spark_df = materialized_spark_df.filter(
            materialized_spark_df[fd.timestamp_key] >= feature_start_time
        )
        materialized_spark_df = materialized_spark_df.filter(materialized_spark_df[fd.timestamp_key] < feature_end_time)

    return tecton.interactive.data_frame.DataFrame._create(materialized_spark_df)


def run_stream(fd: FeatureDefinition, output_temp_table: str) -> StreamingQuery:
    plan = materialization_plan.get_stream_materialization_plan(
        TectonContext.get_instance()._spark,
        data_sources=fd.data_sources,
        transformations=fd.transformations,
        feature_definition=fd,
    )
    spark_df = plan.online_store_data_frame
    with tempfile.TemporaryDirectory() as d:
        return (
            spark_df.writeStream.format("memory")
            .queryName(output_temp_table)
            .option("checkpointLocation", d)
            .outputMode("append")
            .start()
        )


def run_ondemand(
    fd: FeatureDefinition, fv_name: str, mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame, DataFrame]]
) -> Union[Dict[str, Any], "tecton.interactive.data_frame.DataFrame"]:
    fv_proto = fd.feature_view_proto
    for key in mock_inputs:
        if isinstance(mock_inputs[key], DataFrame):
            mock_inputs[key] = mock_inputs[key].toPandas()

    # Validate that all the mock_inputs matchs with FV inputs, and that num rows match across all mock_inputs.
    validate_ondemand_mock_inputs_keys(mock_inputs, fv_proto)

    # Execute ODFV pipeline to get output DataFrame.
    output = run_mock_odfv_pipeline(
        pipeline=fd.pipeline,
        transformations=fd.transformations,
        name=fv_name,
        mock_inputs=mock_inputs,
    )
    if isinstance(output, pandas.DataFrame):
        output = tecton.interactive.data_frame.DataFrame._create(output)

    return output


def validate_and_get_aggregation_level(fd: FeatureDefinition, aggregate_tiles: bool, aggregation_level: str) -> str:
    if aggregate_tiles is not None:
        if aggregation_level:
            raise errors.FV_INVALID_ARG_COMBO(["aggregate_tiles", "aggregation_level"])
        logger.warning("aggregate_tiles is deprecated - please use aggregation_level instead.")
        aggregation_level = AGGREGATION_LEVEL_PARTIAL if aggregate_tiles else AGGREGATION_LEVEL_DISABLED

    # Set default aggregation_level value.
    if aggregation_level is None:
        if fd.is_temporal_aggregate and not fd.trailing_time_window_aggregation.is_continuous:
            aggregation_level = AGGREGATION_LEVEL_FULL
        else:
            aggregation_level = AGGREGATION_LEVEL_DISABLED

    if aggregation_level not in SUPPORTED_AGGREGATION_LEVEL_VALUES:
        raise errors.FV_INVALID_ARG_VALUE(
            "aggregation_level", str(aggregation_level), str(SUPPORTED_AGGREGATION_LEVEL_VALUES)
        )

    if aggregation_level != "disabled" and fd.trailing_time_window_aggregation.is_continuous:
        raise errors.TectonValidationError(
            "aggregate_level is not supported for this FeatureView, since aggregation_slide_period='continuous'."
        )
    return aggregation_level


# For single-batch-schedule-interval-per-job backfill, validate the followings.
# - Only support single-tile run.
# - Don't allow passing `feature_start_time` without feature_end_time since it may be confusing that the tile time
#   range goes into the future.
def _validate_feature_time_for_backfill_config(
    fv_proto: feature_view_pb2.FeatureView,
    feature_start_time: Optional[Union[pendulum.DateTime, datetime]],
    feature_end_time: Optional[Union[pendulum.DateTime, datetime]],
    feature_time_limits_aligned: pendulum.Period,
):
    # TODO(raviphol): Use utils.is_bfc_mode_single once D9614 is landed.
    if not fv_proto.HasField("temporal"):
        return
    if not fv_proto.temporal.HasField("backfill_config"):
        return
    backfill_config_mode = fv_proto.temporal.backfill_config.mode
    if backfill_config_mode is not BackfillConfigMode.BACKFILL_CONFIG_MODE_SINGLE_BATCH_SCHEDULE_INTERVAL_PER_JOB:
        return

    if feature_start_time and not feature_end_time:
        raise errors.BFC_MODE_SINGLE_REQUIRED_FEATURE_END_TIME_WHEN_START_TIME_SET

    schedule_interval_seconds = fv_proto.materialization_params.schedule_interval.ToSeconds()
    if schedule_interval_seconds == 0:
        raise errors.INTERNAL_ERROR("Materialization schedule interval not found.")

    num_tile = feature_time_limits_aligned.in_seconds() // schedule_interval_seconds
    if num_tile > 1:
        raise errors.BFC_MODE_SINGLE_INVALID_FEATURE_TIME_RANGE


# Validate that mock_inputs keys are a subset of data sources.
def _validate_batch_mock_inputs_keys(mock_inputs, fv_proto):
    expected_input_names = get_all_input_keys(fv_proto.pipeline.root)
    mock_inputs_keys = set(mock_inputs.keys())
    if not mock_inputs_keys.issubset(expected_input_names):
        raise errors.FV_INVALID_MOCK_INPUTS(mock_inputs_keys, expected_input_names)


# Validate that mock_inputs keys are exact match with expected inputs.
def validate_ondemand_mock_inputs_keys(
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame]],
    fv_proto: feature_view_pb2.FeatureView,
):
    expected_input_names = get_all_input_keys(fv_proto.pipeline.root)
    mock_inputs_keys = set(mock_inputs.keys())
    if mock_inputs_keys != expected_input_names:
        raise errors.FV_INVALID_MOCK_INPUTS(mock_inputs_keys, expected_input_names)

    # Get num row for all FV mock_inputs with DF types, to validate that they match.
    input_df_row_counts = set()
    for input in mock_inputs.values():
        if isinstance(input, pandas.DataFrame):
            input_df_row_counts.add(len(input.index))
    if len(input_df_row_counts) > 1:
        raise errors.FV_INVALID_MOCK_INPUTS_NUM_ROWS(input_df_row_counts)


# Check that schema of each mock inputs matches with data sources.
def _validate_input_dataframe_schema(input_name, dataframe: DataFrame, spark_schema):
    columns = sorted(dataframe.columns)
    expected_column_names = sorted([field.name for field in spark_schema.fields])

    # Validate mock input's schema against expected schema.
    if not expected_column_names == columns:
        raise errors.FV_INVALID_MOCK_INPUT_SCHEMA(input_name, columns, expected_column_names)


def _get_ds_by_id(data_sources, id: str):
    for ds in data_sources:
        if IdHelper.to_string(ds.virtual_data_source_id) == id:
            return ds
    return None


# Align feature start and end times with materialization schedule interval.
def _align_times(
    feature_start_time: datetime,
    feature_end_time: datetime,
    fv_proto: feature_view_pb2.FeatureView,
    schedule_interval: timedelta,
) -> pendulum.Period:
    # Align feature_end_time upward to the nearest materialization schedule interval.
    feature_end_time = time_utils.align_time_upwards(feature_end_time, schedule_interval)

    # Align feature_start_time downward to the nearest materialization schedule interval.
    feature_start_time = time_utils.align_time_downwards(feature_start_time, schedule_interval)
    return pendulum.period(feature_start_time, feature_end_time)


def _get_spark_schema(ds):
    if ds.HasField("batch_data_source"):
        spark_schema = ds.batch_data_source.spark_schema
    elif ds.HasField("stream_data_source"):
        spark_schema = ds.stream_data_source.spark_schema
    else:
        raise errors.INTERNAL_ERROR("DataSource is missing a supporting config")
    return SparkSchemaWrapper.from_proto(spark_schema).unwrap()



from typing import List
from typing import Optional
from typing import Union

import tecton
from tecton import conf
from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.sdk_decorators import documented_by
from tecton._internals.sdk_decorators import sdk_public_method
from tecton._internals.utils import is_live_workspace
from tecton.interactive.dataset import Dataset
from tecton.interactive.feature_table import FeatureTable
from tecton.interactive.feature_view import FeatureView
from tecton.interactive.transformation import Transformation
from tecton_proto.metadataservice.metadata_service_pb2 import ListWorkspacesRequest
from tecton_spark.logger import get_logger

logger = get_logger("Workspace")


class Workspace:
    """
    Workspace class.

    This class represents a Workspace. The Workspace class is used to fetch Tecton Objects, which are stored in a Workspace.

    Examples of using the Workspace methods for accessing Tecton first class objects:

    .. code-block:: python

        import tecton


        WORKSPACE_NAME = "jsd_tecton_wksp"

        workspace = tecton.Workspace(WORKSPACE_NAME)

        # For a specified workspace, list high-level registered Objects
        print(f" Entities            : {workspace.list_entities()}")
        print(f" Feature Data Sources: {workspace.list_data_sources()}")
        print(f" Feature Views       : {workspace.list_feature_views()}")
        print(f" Feature Services    : {workspace.list_feature_services()}")
        print(f" Transformations     : {workspace.list_transformations()}")
        print(f" Feature Tables      : {workspace.list_feature_tables()}")
        print(f" List of Workspaces  : {tecton.list_workspaces()}")

    .. code-block:: Text

        The output:

        Entities            : ['ad', 'auction', 'content', 'ContentKeyword', 'ads_user', 'fraud_user']
        Feature Data Sources: ['ad_impressions_stream', 'ad_impressions_batch', 'transactions_stream', 'transactions_batch',
                                'users_batch', 'credit_scores_batch']
        Feature Views       : ['user_has_great_credit', 'user_age', 'transaction_amount_is_high',
                               'transaction_amount_is_higher_than_average', 'transaction_bucketing', 'user_ctr_7d_2',
                               ...
                               'user_ad_impression_counts', 'content_keyword_click_counts']
        Feature Services    : ['ad_ctr_feature_service', 'fraud_detection_feature_service',
                                'fraud_detection_feature_service:v2', 'minimal_fs', 'continuous_feature_service']
        Transformations     : ['content_keyword_click_counts', 'user_ad_impression_counts', 'user_click_counts',
                               'user_impression_counts', 'user_ctr_7d', 'user_ctr_7d_2', 'user_distinct_ad_count_7d',
                               ...
                               'weekend_transaction_count_n_days', 'user_has_good_credit_sql']
        Feature Tables      : ['user_login_counts', 'user_page_click_feature_table']
        List of Workspaces  : ['jsd_tecton_wksp, 'kafka_streaming_staging, 'kafka_streaming_production',
                               ...
                               'on_demand_streaming_aggregation_pipeline']

    """

    def __init__(self, workspace: str, _is_live: Optional[bool] = None, _validate: bool = True):
        """
        Fetch an existing :class:`Workspace` by name.

        :param workspace: Workspace name.
        """
        self.workspace = workspace

        if _is_live is None:
            self.is_live = is_live_workspace(self.workspace)
        else:
            self.is_live = _is_live

        if _validate:
            self._validate()

    def _validate(self):
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)

        workspace_from_resp = None
        for ws in response.workspaces:
            if ws.name == self.workspace:
                workspace_from_resp = ws
                break

        if workspace_from_resp is None:
            raise errors.NONEXISTENT_WORKSPACE(workspace, workspaces)

        if ws.capabilities.materializable != self.is_live:
            raise errors.INCORRECT_MATERIALIZATION_ENABLED_FLAG(self.is_live, ws.capabilities.materializable)

    @classmethod
    @sdk_public_method
    def get_all(self) -> List["Workspace"]:
        """
        Returns a list of all registered Workspaces.

        :return: A list of Workspace objects.
        """
        request = ListWorkspacesRequest()
        response = metadata_service.instance().ListWorkspaces(request)
        workspaces = [
            Workspace(ws.name, _is_live=ws.capabilities.materializable, _validate=False) for ws in response.workspaces
        ]

        # Return live workspaces first (alphabetical), then development workspaces.
        return sorted(workspaces, key=lambda ws: (not ws.is_live, ws.workspace))

    def __enter__(self):
        logger.warning("`with Workspace('<name>')` is deprecated. Please use Workspace object methods directly.")

        self.previous_workspace = conf.get_or_none("TECTON_WORKSPACE")
        conf.set("TECTON_WORKSPACE", self.workspace)

    def __exit__(self, type, value, traceback):
        conf.set("TECTON_WORKSPACE", self.previous_workspace)

    def __repr__(self) -> str:
        capability_str = "Live" if self.is_live else "Development"
        return f"{self.workspace} ({capability_str})"

    @sdk_public_method
    def summary(self) -> Displayable:
        from texttable import Texttable

        items = [
            ("Workspace Name", self.workspace),
            ("Workspace Type", "Live" if self.is_live else "Development"),
        ]
        return Displayable.from_items(
            headings=["", ""], items=items, deco=(Texttable.BORDER | Texttable.VLINES | Texttable.HLINES)
        )

    @classmethod
    @sdk_public_method
    def get(cls, name) -> "Workspace":
        """
        Fetch an existing :class:`Workspace` by name.

        :param name: Workspace name.
        """
        return Workspace(name)

    @sdk_public_method
    def get_feature_view(self, name: str) -> FeatureView:
        """
        Returns a FeatureView within a workspace.

        :param name: FeatureView name
        :return: :class:`BatchFeatureView`, :class:`BatchWindowAggregateFeatureView`,
            :class:`StreamFeatureView`, :class:`StreamWindowAggregateFeatureView`,
            or :class:`OnDemandFeatureView`
        """
        return tecton.get_feature_view(name, workspace_name=self.workspace)

    @sdk_public_method
    def get_feature_table(self, name: str) -> FeatureTable:
        """
        Returns a :class:`FeatureTable` within a workspace.

        :param name: FeatureTable name
        :return: the named FeatureTable
        """
        return tecton.get_feature_table(name, workspace_name=self.workspace)

    @sdk_public_method
    def get_feature_service(self, name: str):
        """
        Returns a :class:`FeatureService` within a workspace.

        :param name: FeatureService name.
        :return: the named FeatureService
        """

        return tecton.get_feature_service(name, workspace_name=self.workspace)

    @sdk_public_method
    def get_data_source(self, name: str):
        """
        Returns a :class:`BatchDataSource` or :class:`StreamDataSource` within a workspace.

        :param name: BatchDataSource or StreamDataSource name.
        :return: the named BatchDataSource or StreamDataSource
        """

        return tecton.get_data_source(name, workspace_name=self.workspace)

    @sdk_public_method
    def get_entity(self, name: str):
        """
        Returns an :class:`Entity` within a workspace.

        :param name: Entity name.
        :return: the named Entity
        """

        return tecton.get_entity(name, workspace_name=self.workspace)

    @sdk_public_method
    def get_transformation(self, name: str) -> Union[Transformation]:
        """
        Returns a :class:`Transformation` within a workspace.

        :param name: Transformation name.
        :return: the named Transformation
        """

        return tecton.get_transformation(name, workspace_name=self.workspace)

    @sdk_public_method
    def get_dataset(self, name) -> Dataset:
        """
        Returns a :class:`Dataset` within a workspace.

        :param name: Dataset name.
        :return: the named Dataset
        """
        return tecton.get_dataset(name, workspace_name=self.workspace)

    @sdk_public_method
    def list_datasets(self) -> List[str]:
        """
        Returns a list of all registered Datasets within a workspace.

        :return: A list of strings.
        """
        return tecton.list_datasets(workspace_name=self.workspace)

    @sdk_public_method
    def list_feature_views(self) -> List[str]:
        """
        Returns a list of all registered FeatureViews within a workspace.

        :return: A list of strings.
        """
        return tecton.list_feature_views(workspace_name=self.workspace)

    @sdk_public_method
    def list_feature_services(self) -> List[str]:
        """
        Returns a list of all registered FeatureServices within a workspace.

        :return: A list of strings.
        """
        return tecton.list_feature_services(workspace_name=self.workspace)

    @sdk_public_method
    def list_transformations(self) -> List[str]:
        """
        Returns a list of all registered Transformations within a workspace.

        :return: A list of strings.
        """
        return tecton.list_transformations(workspace_name=self.workspace)

    @sdk_public_method
    def list_entities(self) -> List[str]:
        """
        Returns a list of all registered Entities within a workspace.

        :returns: A list of strings.
        """
        return tecton.list_entities(workspace_name=self.workspace)

    @sdk_public_method
    def list_data_sources(self) -> List[str]:
        """
        Returns a list of all registered DataSources within a workspace.

        :return: A list of strings.
        """
        return tecton.list_data_sources(workspace_name=self.workspace)

    @sdk_public_method
    def list_feature_tables(self) -> List[str]:
        """
        Returns a list of all registered FeatureTables within a workspace.

        :return: A list of strings.
        """
        return tecton.list_feature_tables(workspace_name=self.workspace)


@documented_by(Workspace.get)
@sdk_public_method
def get_workspace(name: str):
    return Workspace.get(name)



import datetime
from typing import List
from typing import Optional

from pyspark.sql.functions import udf
from pyspark.sql.types import ArrayType
from pyspark.sql.types import TimestampType

from tecton_spark.errors import TectonValidationError
from tecton_spark.time_utils import align_time_downwards
from tecton_spark.time_utils import assert_valid_time_string
from tecton_spark.time_utils import strict_pytimeparse
from tecton_spark.time_utils import WINDOW_UNBOUNDED_PRECEDING


def _tecton_sliding_window_impl(
    timestamp: datetime.datetime,
    window_size: str,
    slide_interval: str,
    feature_start: datetime.datetime,
    feature_end: datetime.datetime,
) -> List[datetime.datetime]:
    window_size_td, slide_interval_td = _validate_sliding_window_duration(window_size, slide_interval)

    aligned_feature_start = align_time_downwards(feature_start, slide_interval_td)
    earliest_possible_window_start = align_time_downwards(timestamp, slide_interval_td)
    window_end_cursor = max(aligned_feature_start, earliest_possible_window_start) + slide_interval_td
    windows = []
    # confirm windows are aligned
    while window_end_cursor <= feature_end:
        ts_after_window_start = window_size_td is None or timestamp >= window_end_cursor - window_size_td
        ts_before_window_end = timestamp < window_end_cursor
        if ts_after_window_start and ts_before_window_end:
            windows.append(window_end_cursor - datetime.timedelta(microseconds=1))
            window_end_cursor = window_end_cursor + slide_interval_td
        else:
            break
    return windows


def _validate_and_parse_time(duration: str, field_name: str, allow_unbounded: bool) -> Optional[datetime.timedelta]:
    if allow_unbounded and duration.lower() == WINDOW_UNBOUNDED_PRECEDING:
        duration_td = None
    else:
        # called for nice error message
        assert_valid_time_string(duration, allow_unbounded=allow_unbounded)
        duration_td = datetime.timedelta(seconds=strict_pytimeparse(duration))
    if duration_td is not None and duration_td.total_seconds() <= 0:
        raise TectonValidationError(f"Duration {duration} provided for field {field_name} must be positive.")
    return duration_td


def _validate_sliding_window_duration(
    window_size: str, slide_interval: str
) -> (Optional[datetime.timedelta], datetime.timedelta):
    slide_interval_td = _validate_and_parse_time(slide_interval, "slide_interval", allow_unbounded=False)
    window_size_td = _validate_and_parse_time(window_size, "window_size", allow_unbounded=True)
    if window_size_td is not None:
        # note this also confirms window >= slide since a>0, b>0, a % b = 0 implies a >= b
        if window_size_td.total_seconds() % slide_interval_td.total_seconds() != 0:
            raise TectonValidationError(
                f"Window size {window_size} must be a multiple of slide interval {slide_interval}"
            )
    return window_size_td, slide_interval_td


tecton_sliding_window_udf = udf(_tecton_sliding_window_impl, ArrayType(TimestampType()))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/online_store/status_entry.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/online_store/status_entry.proto',
  package='tecton_proto.online_store',
  syntax='proto2',
  serialized_options=b'\n\026com.tecton.onlinestoreP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,tecton_proto/online_store/status_entry.proto\x12\x19tecton_proto.online_store\"z\n\x0bStatusEntry\x12\x1f\n\x0bsource_type\x18\x01 \x01(\tR\nsourceType\x12)\n\x11raw_data_end_time\x18\x02 \x01(\x03R\x0erawDataEndTime\x12\x1f\n\x0b\x61nchor_time\x18\x03 \x01(\x03R\nanchorTimeB\x1a\n\x16\x63om.tecton.onlinestoreP\x01'
)




_STATUSENTRY = _descriptor.Descriptor(
  name='StatusEntry',
  full_name='tecton_proto.online_store.StatusEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_type', full_name='tecton_proto.online_store.StatusEntry.source_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_data_end_time', full_name='tecton_proto.online_store.StatusEntry.raw_data_end_time', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='rawDataEndTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='anchor_time', full_name='tecton_proto.online_store.StatusEntry.anchor_time', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='anchorTime', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=197,
)

DESCRIPTOR.message_types_by_name['StatusEntry'] = _STATUSENTRY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StatusEntry = _reflection.GeneratedProtocolMessageType('StatusEntry', (_message.Message,), {
  'DESCRIPTOR' : _STATUSENTRY,
  '__module__' : 'tecton_proto.online_store.status_entry_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.online_store.StatusEntry)
  })
_sym_db.RegisterMessage(StatusEntry)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/cli/repo_diff.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.data import state_update_pb2 as tecton__proto_dot_data_dot_state__update__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/cli/repo_diff.proto',
  package='tecton_proto.cli',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n tecton_proto/cli/repo_diff.proto\x12\x10tecton_proto.cli\x1a\x1ctecton_proto/common/id.proto\x1a$tecton_proto/data/state_update.proto\"w\n\x15TectonRepoDiffSummary\x12\x45\n\x0cobject_diffs\x18\x01 \x03(\x0b\x32\".tecton_proto.cli.TectonObjectDiffR\x0bobjectDiffs\x12\x17\n\x07plan_id\x18\x02 \x01(\tR\x06planId\"\xce\x01\n\x10TectonObjectDiff\x12M\n\x0ftransition_type\x18\x01 \x01(\x0e\x32$.tecton_proto.data.FcoTransitionTypeR\x0etransitionType\x12O\n\x0fobject_metadata\x18\x02 \x01(\x0b\x32&.tecton_proto.cli.TectonObjectMetadataR\x0eobjectMetadata\x12\x1a\n\x08warnings\x18\x03 \x03(\tR\x08warnings\"\xa7\x01\n\x14TectonObjectMetadata\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x43\n\x0bobject_type\x18\x02 \x01(\x0e\x32\".tecton_proto.cli.TectonObjectTypeR\nobjectType\x12\x14\n\x05owner\x18\x03 \x01(\tR\x05owner\x12 \n\x0b\x64\x65scription\x18\x04 \x01(\tR\x0b\x64\x65scription*\x8a\x01\n\x10TectonObjectType\x12\x1e\n\x1aTECTON_OBJECT_TYPE_UNKNOWN\x10\x00\x12\n\n\x06\x45NTITY\x10\x01\x12\x0f\n\x0b\x44\x41TA_SOURCE\x10\x02\x12\x12\n\x0eTRANSFORMATION\x10\x03\x12\x10\n\x0c\x46\x45\x41TURE_VIEW\x10\x04\x12\x13\n\x0f\x46\x45\x41TURE_SERVICE\x10\x05'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,tecton__proto_dot_data_dot_state__update__pb2.DESCRIPTOR,])

_TECTONOBJECTTYPE = _descriptor.EnumDescriptor(
  name='TectonObjectType',
  full_name='tecton_proto.cli.TectonObjectType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TECTON_OBJECT_TYPE_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ENTITY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATA_SOURCE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFORMATION', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FEATURE_VIEW', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FEATURE_SERVICE', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=623,
  serialized_end=761,
)
_sym_db.RegisterEnumDescriptor(_TECTONOBJECTTYPE)

TectonObjectType = enum_type_wrapper.EnumTypeWrapper(_TECTONOBJECTTYPE)
TECTON_OBJECT_TYPE_UNKNOWN = 0
ENTITY = 1
DATA_SOURCE = 2
TRANSFORMATION = 3
FEATURE_VIEW = 4
FEATURE_SERVICE = 5



_TECTONREPODIFFSUMMARY = _descriptor.Descriptor(
  name='TectonRepoDiffSummary',
  full_name='tecton_proto.cli.TectonRepoDiffSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='object_diffs', full_name='tecton_proto.cli.TectonRepoDiffSummary.object_diffs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectDiffs', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='plan_id', full_name='tecton_proto.cli.TectonRepoDiffSummary.plan_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='planId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=122,
  serialized_end=241,
)


_TECTONOBJECTDIFF = _descriptor.Descriptor(
  name='TectonObjectDiff',
  full_name='tecton_proto.cli.TectonObjectDiff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='transition_type', full_name='tecton_proto.cli.TectonObjectDiff.transition_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='transitionType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_metadata', full_name='tecton_proto.cli.TectonObjectDiff.object_metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectMetadata', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warnings', full_name='tecton_proto.cli.TectonObjectDiff.warnings', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='warnings', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=244,
  serialized_end=450,
)


_TECTONOBJECTMETADATA = _descriptor.Descriptor(
  name='TectonObjectMetadata',
  full_name='tecton_proto.cli.TectonObjectMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tecton_proto.cli.TectonObjectMetadata.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='object_type', full_name='tecton_proto.cli.TectonObjectMetadata.object_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='objectType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='owner', full_name='tecton_proto.cli.TectonObjectMetadata.owner', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='owner', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='tecton_proto.cli.TectonObjectMetadata.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='description', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=453,
  serialized_end=620,
)

_TECTONREPODIFFSUMMARY.fields_by_name['object_diffs'].message_type = _TECTONOBJECTDIFF
_TECTONOBJECTDIFF.fields_by_name['transition_type'].enum_type = tecton__proto_dot_data_dot_state__update__pb2._FCOTRANSITIONTYPE
_TECTONOBJECTDIFF.fields_by_name['object_metadata'].message_type = _TECTONOBJECTMETADATA
_TECTONOBJECTMETADATA.fields_by_name['object_type'].enum_type = _TECTONOBJECTTYPE
DESCRIPTOR.message_types_by_name['TectonRepoDiffSummary'] = _TECTONREPODIFFSUMMARY
DESCRIPTOR.message_types_by_name['TectonObjectDiff'] = _TECTONOBJECTDIFF
DESCRIPTOR.message_types_by_name['TectonObjectMetadata'] = _TECTONOBJECTMETADATA
DESCRIPTOR.enum_types_by_name['TectonObjectType'] = _TECTONOBJECTTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TectonRepoDiffSummary = _reflection.GeneratedProtocolMessageType('TectonRepoDiffSummary', (_message.Message,), {
  'DESCRIPTOR' : _TECTONREPODIFFSUMMARY,
  '__module__' : 'tecton_proto.cli.repo_diff_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.cli.TectonRepoDiffSummary)
  })
_sym_db.RegisterMessage(TectonRepoDiffSummary)

TectonObjectDiff = _reflection.GeneratedProtocolMessageType('TectonObjectDiff', (_message.Message,), {
  'DESCRIPTOR' : _TECTONOBJECTDIFF,
  '__module__' : 'tecton_proto.cli.repo_diff_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.cli.TectonObjectDiff)
  })
_sym_db.RegisterMessage(TectonObjectDiff)

TectonObjectMetadata = _reflection.GeneratedProtocolMessageType('TectonObjectMetadata', (_message.Message,), {
  'DESCRIPTOR' : _TECTONOBJECTMETADATA,
  '__module__' : 'tecton_proto.cli.repo_diff_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.cli.TectonObjectMetadata)
  })
_sym_db.RegisterMessage(TectonObjectMetadata)


# @@protoc_insertion_point(module_scope)



import json
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Iterable
from typing import Optional

from tecton_spark import errors

logging.getLogger("boto3").setLevel(logging.ERROR)
logging.getLogger("botocore").setLevel(logging.ERROR)

_CONFIG_OVERRIDES = {}


class _Debugger(object):
    class ConfSource(Enum):
        SESSION_OVERRIDE = 1
        OS_ENV = 2
        LOCAL_TECTON_CONFIG = 3
        REMOTE_MDS_CONFIG = 4
        RUNTIME_CONFIG = 5
        DATABRICKS_SECRET = 6
        AWS_SECRET_MANAGER = 7
        DEFAULT = 8
        NOT_FOUND = 9

    @classmethod
    def _debug_enabled(cls) -> bool:
        key = "TECTON_DEBUG"
        if key in os.environ:
            return os.environ[key] == "1"
        elif key in _CONFIG_OVERRIDES:
            return _CONFIG_OVERRIDES[key] == "1"
        return False

    @classmethod
    def print(cls, src: ConfSource, key: str, val=None):
        if not cls._debug_enabled():
            return

        print("🔍", f"Looking up {key}")
        indent = ""
        for s in cls.ConfSource:
            if src == s:
                print(indent, "✅", f"{s.name} -> {val}")
                break
            print(indent, "👎", s.name)
            indent += " "


def set(key, value):
    _set(key, value)


def _set(key, value):
    _CONFIG_OVERRIDES[key] = value


def _does_key_have_valid_prefix(key) -> bool:
    for prefix in _VALID_KEY_PREFIXES:
        if key.startswith(prefix):
            return True
    return False


def _get(key) -> Optional[str]:
    """Get the config value for the given key, or return None if not found."""
    if key not in _VALID_KEYS and not _does_key_have_valid_prefix(key):
        raise errors.TectonInternalError(f"Tried accessing invalid configuration key '{key}'")

    # Session-scoped override.
    if key in _CONFIG_OVERRIDES:
        val = _CONFIG_OVERRIDES[key]
        _Debugger.print(_Debugger.ConfSource.SESSION_OVERRIDE, key, val)
        return val

    # Environment variable.
    if key in os.environ:
        val = os.environ[key]
        _Debugger.print(_Debugger.ConfSource.OS_ENV, key, val)
        return val

    # ~/.tecton/config
    if key in _LOCAL_TECTON_CONFIG:
        val = _LOCAL_TECTON_CONFIG[key]
        _Debugger.print(_Debugger.ConfSource.LOCAL_TECTON_CONFIG, key, val)
        return val

    # Config from MDS
    if key in _REMOTE_MDS_CONFIGS:
        val = _REMOTE_MDS_CONFIGS[key]
        _Debugger.print(_Debugger.ConfSource.REMOTE_MDS_CONFIG, key, val)
        return val

    # Compute runtime configs.
    if key in _RUNTIME_CONFIGS:
        val = _RUNTIME_CONFIGS[key]
        _Debugger.print(_Debugger.ConfSource.RUNTIME_CONFIG, key, val)
        return val

    if _get_runtime_env() == TectonEnv.UNKNOWN:
        # Fallback attempt to set env if user has not set it.
        _set_tecton_runtime_env()

    if _should_lookup_db_secrets(key):
        # Databricks secrets
        for scope in _get_secret_scopes():
            value = _get_from_db_secrets(key, scope)
            if value:
                _Debugger.print(_Debugger.ConfSource.DATABRICKS_SECRET, key, value)
                return value

    if _should_lookup_aws_secretsmanager(key):
        # AWS secret manager
        for scope in _get_secret_scopes():
            value = _get_from_secretsmanager(key, scope)
            if value:
                _Debugger.print(_Debugger.ConfSource.AWS_SECRET_MANAGER, key, value)
                return value

    if key in _DEFAULTS:
        value = _DEFAULTS[key]()
        _Debugger.print(_Debugger.ConfSource.DEFAULT, key, value)
        return value

    _Debugger.print(_Debugger.ConfSource.NOT_FOUND, key)
    return None


def get_or_none(key) -> Optional[str]:
    return _get(key)


def get_or_raise(key) -> str:
    val = _get(key)
    if val == None:
        raise errors.TectonInternalError(f"{key} not set")
    return val


# Internal

_LOCAL_TECTON_CONFIG_FILE = Path(os.environ.get("TECTON_CONFIG_PATH", Path.home() / ".tecton/config"))
_LOCAL_TECTON_TOKENS_FILE = _LOCAL_TECTON_CONFIG_FILE.with_suffix(".tokens")

# Keys stored in the tecton config file
_TECTON_CONFIG_FILE_KEYS = {
    "API_SERVICE",
    "FEATURE_SERVICE",
    "TECTON_WORKSPACE",
    "CLI_CLIENT_ID",
    "SNOWFLAKE_USER",
    "SNOWFLAKE_WAREHOUSE",
    "ALPHA_SNOWFLAKE_COMPUTE_ENABLED",
}

# Keys stored in the tecton tokens file
_TECTON_TOKENS_FILE_KEYS = {
    "OAUTH_ACCESS_TOKEN",
    "OAUTH_ACCESS_TOKEN_EXPIRATION",
    "OAUTH_REFRESH_TOKEN",
    "SNOWFLAKE_PASSWORD",
}

_VALID_KEYS = [
    # Alpha features
    "ALPHA_SNOWFLAKE_COMPUTE_ENABLED",
    "ALPHA_SNOWFLAKE_SNOWPARK_ENABLED",
    "API_SERVICE",
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "FEATURE_SERVICE",
    "HIVE_METASTORE_HOST",
    "HIVE_METASTORE_PORT",
    "HIVE_METASTORE_USERNAME",
    "HIVE_METASTORE_DATABASE",
    "HIVE_METASTORE_PASSWORD",
    "SPARK_DRIVER_LOCAL_IP",
    "METADATA_SERVICE",
    "TECTON_CLUSTER_NAME",
    "TECTON_API_KEY",
    "OAUTH_ACCESS_TOKEN",
    "OAUTH_ACCESS_TOKEN_EXPIRATION",
    "OAUTH_REFRESH_TOKEN",
    "CLI_CLIENT_ID",
    "USE_DIRECT_GRPC",
    "TECTON_WORKSPACE",
    "CLUSTER_REGION",
    "REDSHIFT_USER",
    "REDSHIFT_PASSWORD",
    "SNOWFLAKE_ACCOUNT_IDENTIFIER",
    "SNOWFLAKE_USER",
    "SNOWFLAKE_PASSWORD",
    "SNOWFLAKE_WAREHOUSE",
    "SNOWFLAKE_DATABASE",
    "TECTON_DEBUG",
    "TECTON_RUNTIME_ENV",
    "TECTON_RUNTIME_MODE",
]

_VALID_KEY_PREFIXES = ["SECRET_"]

_DEFAULTS = {
    "TECTON_WORKSPACE": (lambda: "prod"),
    "FEATURE_SERVICE": (lambda: _get("API_SERVICE")),
    "ALPHA_SNOWFLAKE_SNOWPARK_ENABLED": (lambda: True),
}

_REMOTE_MDS_CONFIGS = {}

_RUNTIME_CONFIGS = {"": ""}

_RUNTIME_CONFIGS_VALID_KEYS = [
    "TECTON_RUNTIME_ENV",
    "TECTON_RUNTIME_MODE",
]

_is_running_on_databricks_cache = None
_is_running_on_emr_cache = None
TectonEnv = Enum("TectonEnv", "DATABRICKS EMR UNKNOWN")


def _is_running_on_databricks():
    """Whether we're running in Databricks notebook or not."""
    global _is_running_on_databricks_cache
    if _is_running_on_databricks_cache is None:
        main = __import__("__main__")
        filename = os.path.basename(getattr(main, "__file__", ""))
        is_python_shell = filename == "PythonShell.py"
        is_databricks_env = "DBUtils" in main.__dict__
        _is_running_on_databricks_cache = is_python_shell and is_databricks_env
    return _is_running_on_databricks_cache


def _is_running_on_emr():
    """Whether we're running in EMR notebook or not."""
    global _is_running_on_emr_cache
    if _is_running_on_emr_cache is None:
        _is_running_on_emr_cache = "EMR_CLUSTER_ID" in os.environ
    return _is_running_on_emr_cache


def _set_tecton_runtime_env():
    key = "TECTON_RUNTIME_ENV"
    if _is_running_on_databricks():
        set_runtime_config(key, "DATABRICKS")
    elif _is_running_on_emr():
        set_runtime_config(key, "EMR")
    else:
        set_runtime_config(key, "UNKNOWN")


def _is_mode_materialization():
    key = "TECTON_RUNTIME_MODE"
    if key in _RUNTIME_CONFIGS and _RUNTIME_CONFIGS[key] == "MATERIALIZATION":
        return True
    if key in os.environ and os.environ[key] == "MATERIALIZATION":
        return True
    return False


def _get_runtime_env():
    key = "TECTON_RUNTIME_ENV"
    if key in _RUNTIME_CONFIGS and _RUNTIME_CONFIGS[key] == "DATABRICKS":
        return TectonEnv.DATABRICKS
    if key in os.environ and os.environ[key] == "DATABRICKS":
        return TectonEnv.DATABRICKS
    if key in _RUNTIME_CONFIGS and _RUNTIME_CONFIGS[key] == "EMR":
        return TectonEnv.EMR
    if key in os.environ and os.environ[key] == "EMR":
        return TectonEnv.EMR
    return TectonEnv.UNKNOWN


def set_runtime_config(key: str, value: str):
    if key not in _RUNTIME_CONFIGS_VALID_KEYS:
        raise errors.TectonInternalError(f"Tried accessing invalid configuration key '{key}'")
    _RUNTIME_CONFIGS[key] = value


def _should_lookup_aws_secretsmanager(key: str) -> bool:
    # Keys used for secret manager lookups that cause infinite loops.
    return key not in ("CLUSTER_REGION", "TECTON_CLUSTER_NAME", "USE_DIRECT_GRPC")


def _should_lookup_db_secrets(key: str) -> bool:
    return _get_runtime_env() != TectonEnv.EMR and key not in ("TECTON_CLUSTER_NAME")


def _get_dbutils():
    # Returns dbutils import. Only works in Databricks notebook environment
    import IPython

    return IPython.get_ipython().user_ns["dbutils"]


def save_tecton_configs():
    _save_tecton_config(_LOCAL_TECTON_CONFIG_FILE, _TECTON_CONFIG_FILE_KEYS)
    _save_tecton_config(_LOCAL_TECTON_TOKENS_FILE, _TECTON_TOKENS_FILE_KEYS)


def _save_tecton_config(path: Path, keys: Iterable[str]):
    tecton_config = {key: get_or_none(key) for key in keys if get_or_none(key) is not None}
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as f:
        json.dump(tecton_config, f, sort_keys=True, indent=2)
        f.write("\n")


# Get key by looking in TECTON_CLUSTER_NAME'd scope and falling back to "tecton"
def _get_secret_scopes():
    cluster_name = get_or_none("TECTON_CLUSTER_NAME")
    secret_scopes = []
    if cluster_name:
        secret_prefix = cluster_name if cluster_name.startswith("tecton-") else f"tecton-{cluster_name}"
        secret_scopes.append(secret_prefix)
    secret_scopes.append("tecton")
    return secret_scopes


def _get_from_secretsmanager(key: str, scope: str):
    try:
        # Try to Grab secret from AWS secrets manager
        from boto3 import client

        if _is_mode_materialization():
            aws_secret_client = client("secretsmanager")
        else:
            aws_secret_client = client("secretsmanager", region_name=get_or_none("CLUSTER_REGION"))
        secret = aws_secret_client.get_secret_value(SecretId=f"{scope}/{key}")
        return secret["SecretString"]
    except Exception:
        # Do not fail if secret is not found
        return None


def _get_from_db_secrets(key: str, scope: str):
    try:
        dbutils = _get_dbutils()
        return dbutils.secrets.get(scope, key)
    except Exception:
        return None


def save_okta_tokens(access_token, access_token_expiration, refresh_token=None):
    _set("OAUTH_ACCESS_TOKEN", access_token)
    _set("OAUTH_ACCESS_TOKEN_EXPIRATION", access_token_expiration)
    if refresh_token:
        _set("OAUTH_REFRESH_TOKEN", refresh_token)
    _save_tecton_config(_LOCAL_TECTON_TOKENS_FILE, _TECTON_TOKENS_FILE_KEYS)


def _read_json_config(file_path: Path):
    """If the file exists, reads it and returns parsed JSON. Otherwise returns empty dictionary."""
    if not file_path.exists():
        return {}
    content = file_path.read_text()
    if not content:
        return {}
    try:
        return json.loads(content)
    except json.decoder.JSONDecodeError as e:
        raise ValueError(
            f"Unable to decode JSON configuration file {file_path} ({str(e)}). "
            + "To regenerate configuration, delete this file and run `tecton login`."
        )


_LOCAL_TECTON_CONFIG = None


def _init_configs():
    if not _is_mode_materialization():
        global _LOCAL_TECTON_CONFIG
        _LOCAL_TECTON_CONFIG = _read_json_config(_LOCAL_TECTON_CONFIG_FILE)
        for key, val in _read_json_config(_LOCAL_TECTON_TOKENS_FILE).items():
            _LOCAL_TECTON_CONFIG[key] = val


def _init_metadata_server_config(mds_response):
    global _REMOTE_MDS_CONFIGS
    _REMOTE_MDS_CONFIGS = dict(mds_response.key_values)


_init_configs()



from typing import Dict
from typing import Optional

from tecton_proto.args import basic_info_pb2


def prepare_basic_info(
    *,
    name: str,
    description: Optional[str],
    family: Optional[str],
    tags: Optional[Dict[str, str]],
    owner: Optional[str]
):
    info = basic_info_pb2.BasicInfo()
    info.name = name
    info.description = description or ""
    info.family = family or ""
    info.tags.update(tags or {})
    info.owner = owner or ""
    return info



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/fco_locator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/common/fco_locator.proto',
  package='tecton_proto.common',
  syntax='proto2',
  serialized_options=b'\n\021com.tecton.commonP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%tecton_proto/common/fco_locator.proto\x12\x13tecton_proto.common\x1a\x1ctecton_proto/common/id.proto\"v\n\nFcoLocator\x12\x14\n\x04name\x18\x01 \x01(\tH\x00R\x04name\x12)\n\x02id\x18\x02 \x01(\x0b\x32\x17.tecton_proto.common.IdH\x00R\x02id\x12\x1c\n\tworkspace\x18\x03 \x01(\tR\tworkspaceB\t\n\x07locatorB\x15\n\x11\x63om.tecton.commonP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])




_FCOLOCATOR = _descriptor.Descriptor(
  name='FcoLocator',
  full_name='tecton_proto.common.FcoLocator',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tecton_proto.common.FcoLocator.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='tecton_proto.common.FcoLocator.id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='id', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='workspace', full_name='tecton_proto.common.FcoLocator.workspace', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='workspace', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='locator', full_name='tecton_proto.common.FcoLocator.locator',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=92,
  serialized_end=210,
)

_FCOLOCATOR.fields_by_name['id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FCOLOCATOR.oneofs_by_name['locator'].fields.append(
  _FCOLOCATOR.fields_by_name['name'])
_FCOLOCATOR.fields_by_name['name'].containing_oneof = _FCOLOCATOR.oneofs_by_name['locator']
_FCOLOCATOR.oneofs_by_name['locator'].fields.append(
  _FCOLOCATOR.fields_by_name['id'])
_FCOLOCATOR.fields_by_name['id'].containing_oneof = _FCOLOCATOR.oneofs_by_name['locator']
DESCRIPTOR.message_types_by_name['FcoLocator'] = _FCOLOCATOR
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FcoLocator = _reflection.GeneratedProtocolMessageType('FcoLocator', (_message.Message,), {
  'DESCRIPTOR' : _FCOLOCATOR,
  '__module__' : 'tecton_proto.common.fco_locator_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.common.FcoLocator)
  })
_sym_db.RegisterMessage(FcoLocator)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
